#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import json
import logging
import re
import shlex
import subprocess32
from functools import partial, wraps
from zmon_worker_monitor.zmon_worker.errors import CheckError, NagiosError

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


logger = logging.getLogger(__name__)


# only return 95% of diskspace because of http://unix.stackexchange.com/questions/7950/reserved-space-for-root-on-a-filesystem-why
USABLE_DISKSPACE_FACTOR = 0.95


class NagiosFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(NagiosFactory, self).__init__()

        # fields from config
        self._exarpc_user = None
        self._exarpc_pass = None

        self._loungemysql_user = None
        self._loungemysql_pass = None

        self._hetcrawler_proxy_user = None
        self._hetcrawler_proxy_pass = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self._exarpc_user = conf['exarpc_user']
        self._exarpc_pass = conf['exarpc_pass']

        self._loungemysql_user = conf['loungemysql_user']
        self._loungemysql_pass = conf['loungemysql_pass']

        self._hetcrawler_proxy_user = conf['hetcrawler_proxy_user']
        self._hetcrawler_proxy_pass = conf['hetcrawler_proxy_pass']

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(NagiosWrapper,
                          factory_ctx['host'],
                          exasol_user=self._exarpc_user,
                          exasol_password=self._exarpc_pass,
                          lounge_mysql_user=self._loungemysql_user,
                          lounge_mysql_password=self._loungemysql_pass,
                          hetcrawler_proxy_user=self._hetcrawler_proxy_user,
                          hetcrawler_proxy_pass=self._hetcrawler_proxy_pass)


def error_wrapped(parser):

    @wraps(parser)
    def wrapper(*args, **kwargs):
        try:
            result = parser(*args, **kwargs)
        except Exception:
            raise NagiosError(args[0])
        else:
            return result

    return wrapper


def fix_sub32_exc(e):
    '''Quick and dirty way to deal with subprocess32 bugs. See PF-4190'''

    if isinstance(e, subprocess32.TimeoutExpired):
        for field in ['output', 'timeout', 'cmd']:
            if not hasattr(e, field):
                setattr(e, field, '--')
    if isinstance(e, subprocess32.CalledProcessError):
        for field in ['returncode', 'output', 'cmd']:
            if not hasattr(e, field):
                setattr(e, field, '--')
    return e


class NagiosWrapper(object):

    def __init__(self, host, exasol_user='nagios', exasol_password='', lounge_mysql_user='nagios',
                 lounge_mysql_password='', hetcrawler_proxy_user='', hetcrawler_proxy_pass=''):
        self.host = host
        self.__nrpe_config = {  # example to check non-default memcached:
                              # nagios().nrpe('check_memcachestatus', port=11212)
                              # example to check non-default logwatch:
                              # requires NRPE: command[check_all_disks]=/usr/lib/nagios/plugins/check_disk -w "$ARG1$" -c "$ARG2$" -u "$ARG3$"
            'check_diff_reverse': {'args': '-a /proc/meminfo CommitLimit Committed_AS kB 1048576 524288',
                                   'parser': self._to_dict_commitdiff},
            'check_disk': {'args': '-a 15% 7% /', 'parser': self._parse_memory},
            'check_all_disks': {'args': '-a 15% 7% MB', 'parser': self._parse_disks},
            'check_fiege-avis-file': {'args': '', 'parser': self._to_dict_from_text},
            'check_findfiles': {'args': '-a 20,20,20 20,20,20 {directory} {epoch} found',
                                'parser': partial(self._to_dict, func=int),
                                'parameters': {'directory': '', 'epoch': 1}},
            'check_findfiles_names': {'args': '-a 20,20,20 20,20,20 {directory} {epoch} found {name}',
                                      'parser': partial(self._to_dict, func=int),
                                      'parameters': {'directory': '', 'epoch': 1,
                                      'name': ''}},
            'check_findfiles_names_exclude': {'args': '-a 20,20,20 20,20,20 {directory} {epoch} found {name}',
                                      'parser': partial(self._to_dict, func=int),
                                      'parameters': {'directory': '', 'epoch': 1,
                                      'name': ''}},
            'check_hpacucli_py': {'args': '', 'parser': json.loads},
            'check_hpacucli': {'args': '', 'parser': self._to_dict_hpraid},
            'check_hpasm_dl380p_gen8_fix': {'args': '-a 14:60 15:60', 'parser': self._to_dict_hpasm},
            'check_hpasm_fix_power_supply': {'args': '-a 14:60 15:60', 'parser': self._to_dict_hpasm},
            'check_hpasm_gen8': {'args': '-a 14:60 15:60', 'parser': self._to_dict_hpasm},
            'check_inodes': {'args': '', 'parser': json.loads},
            'check_iostat': {'args': '-a 32000,30000 64000,40000 {disk}', 'parser': self._to_dict_iostat,
                             'parameters': {'disk': 'sda'}},
            'check_list_timeout': {
                'args': '-a "ls {path}" {timeout}',
                'parameters': {'path': '/data/production/', 'timeout': 10},
                'parser': self._to_dict,
                'pre_run_hook': self._check_path_chars,
            },
            'check_load': {'args': '-a 15,14,12 20,17,15', 'parser': self._to_dict},
            'check_mailq_postfix': {'args': '-a 10 5000', 'parser': partial(self._to_dict, func=int)},
            'check_postfix_queue.sh': {'parser': partial(self._to_dict, func=int)},
            'check_memcachestatus': {'args': '-a 9000000,550,10000,100,6000,5000,20481024,20481024 99000000,1000,12000,200,8000,7000,40961024,40961024 127.0.0.1 {port}',
                                     'parser': self._to_dict, 'parameters': {'port': 11211}},
            'check_ntp_time': {'args': '-a 1 2 10 {ntp_server}', 'parser': self._to_dict},
            'check_openmanage': {'args': '', 'parser': self._to_dict_hpasm},
            'check_subdomain_redirect': {'args': '', 'parser': self._to_dict_from_text},
            'check_icmp': {'args': '-a {targethost} {num_of_packets} {timeout}', 'parser': self._to_dict,
                           'parameters': {'targethost': 'default', 'num_of_packets': 5, 'timeout': 10}},
            'check_tcp': {'args': '-a {targethost} {port} {timeout}', 'parser': self._to_dict,
                          'parameters': {'targethost': 'default', 'port': 22, 'timeout': 10}},
            'check_tcp_str': {'args': '-a {targethost} {port} {timeout} {expected}', 'parser': self._to_dict,
                              'parameters': {
                'targethost': 'default',
                'port': 22,
                'timeout': 10,
                'expected': 'SSH-2.0-OpenSSH',
            }},
            'check_ssl': {'args': '-a {targethost} {port} {timeout}', 'parser': self._to_dict,
                          'parameters': {'targethost': 'default', 'port': 443, 'timeout': 10}},
            'check_statistics.pl': {'args': '', 'parser': self._to_dict},
            'check_oracle': {'args': '{user_args}', 'parser': self._to_dict, 'parameters': {'user_args': ''}},
            'check_dbus': {'args': '', 'parser': self._to_dict_win_text,},
            'check_flocked_file': {'args': '-a {lockfile}', 'parser': self._to_dict_from_text},
            'check_apachestatus_uri': {'args': '-a 16000,10000,48 32000,20000,64 {url}', 'parser': self._to_dict,
                                       'parameters': {'url': 'http://127.0.0.1/server-status?auto'}},
            'check_command_procs': {'args': '-a 250 500 {process}', 'parser': self._to_dict_procs,
                                    'parameters': {'process': 'httpd'}},
            'check_http_expect_port_header': {'args': '-a 2 8 60 {ip} {url} {redirect} {size} {expect} {port} {hostname}',
                                              'parser': self._to_dict, 'parameters': {
                'ip': 'localhost',
                'url': '/',
                'redirect': 'warning',
                'size': '9000:90000',
                'expect': '200',
                'port': '88',
                'hostname': 'www.example.com',
            }},
            'check_mysql_processes': {'args': '-a 30 60 {host} {port} {user} {password}',
                                      'parser': self._to_dict_mysql_procs, 'parameters': {
                'host': 'localhost',
                'port': '/var/lib/mysql/mysql.sock',
                'user': lounge_mysql_user,
                'password': lounge_mysql_password,
            }},
            'check_mysqlperformance': {'args': '-a 10000,1500,5000,500,750,100,100,1,5000,30,60,500,10,30 15000,3000,10000,750,1000,250,250,5,7500,60,300,1000,20,60 {host} {port} Questions,Com_select,Qcache_hits,Com_update,Com_insert,Com_delete,Com_replace,Aborted_clients,Com_change_db,Created_tmp_disk_tables,Created_tmp_tables,Qcache_not_cached,Table_locks_waited,Select_scan {user} {password}',
                                       'parser': self._to_dict, 'parameters': {
                'host': 'localhost',
                'port': '/var/lib/mysql/mysql.sock',
                'user': lounge_mysql_user,
                'password': lounge_mysql_password,
            }},
            'check_mysql_slave': {'args': '-a 3 60 {host} {port} {database} {user} {password}',
                                  'parser': self._to_dict_mysql_slave, 'parameters': {
                'host': 'localhost',
                'port': '/var/lib/mysql/mysql.sock',
                'database': 'zlr_live_global',
                'user': lounge_mysql_user,
                'password': lounge_mysql_password,
            }},
            'check_stunnel_target': {
                        'args': '-a {target} {user} {password}',
                        'parser': self._to_dict,
                        'parameters': {
                            'target': 'www.example.com',
                            'user': hetcrawler_proxy_user,
                            'password': hetcrawler_proxy_pass,
                        },
            },
            'check_lounge_queries': {'args': '', 'parser': self._to_dict_lounge_queries},
            'check_newsletter': {'args': '-p {port}', 'parser': self._to_dict_newsletter,
                                 'parameters': {'port': '5666'}},
            'check_nfs_mounts': {'args': '', 'parser': self._to_dict_list},
            'check_kdc': {'args': '', 'parser': json.loads},
            'check_kadmin': {'args': '', 'parser': json.loads},
            'check_ssl_cert': {'args': '-a 60 30 {host_ip} {domain_name}', 'parser': partial(self._to_dict, func=int),
                               'parameters': {'host_ip': '127.0.0.1', 'domain_name': 'www.example.com'}},
        }

        self.__local_config = {
            'check_subdomain_redirect.py': {'args': '', 'parser': self._to_dict_from_text},
            'check_ping': {'args': '-H {} -w 5000,100% -c 5000,100% -p 1'.format(self.host), 'parser': self._to_dict},
            'check_snmp_mem_used-cached.pl': {'args': '-H {} -w 100,100,100 -c 100,100,100 -C public -f'.format(self.host),
                                              'parser': self._to_dict},
            'check_icmp': {'args': '-H {} -n {{num_of_packets}} -t {{timeout}}'.format(self.host),
                           'parser': self._to_dict, 'parameters': {'num_of_packets': 5, 'timeout': 10}},
            'check_tcp': {'args': '-H {} -p {{port}} -t {{timeout}}'.format(self.host), 'parser': self._to_dict,
                          'parameters': {'port': 22, 'timeout': 10}},
            'check_tcp_str': {'args': '-H {} -p {{port}} -t {{timeout}} -e {{expected}}'.format(self.host),
                              'parser': self._to_dict, 'parameters': {'port': 22, 'timeout': 10,
                              'expected': 'SSH-2.0-OpenSSH'}},
            'check_ssl': {'args': '-H {} -p {{port}} -t {{timeout}} -S'.format(self.host), 'parser': self._to_dict,
                          'parameters': {'port': 443, 'timeout': 10}},
            'check_dns': {'args': '-H {host} -s {dns_server} -t {timeout}', 'parser': self._to_dict,
                          'parameters': {'timeout': 5}},
            'check_snmp_process.pl': {'args': '-H {} -C {{community}} -F -n {{name}} -c {{critical}} -w {{warn}} -o {{octets}} {{extra}}'.format(self.host),
                                      'parser': self._to_dict, 'parameters': {
                'timeout': 5,
                'octets': 2400,
                'warn': 1,
                'critical': 1,
                'community': 'public',
                'extra': '-r -2',
            }},
            # check_xmlrpc.rb:  BI checks for PF-3558
            # possible user_args:
            # Exasol: backup state -> '-check-backup'  (default)
            # Exasol: DB-Status -> '--rpc getDatabaseState --ok'
            # Exasol: Node status -> '--check-node-states -w 0 -c 1'
            # Exasol: Verbindungs-Status -> '--rpc getDatabaseConnectionState --ok yes'
            'check_xmlrpc.rb': {'args': '--url http://{user}:{password}@{targethost}/cluster1/db_exa_db1 {user_args}',
                                'parser': self._to_dict_newsletter, 'parameters': {
                'targethost': '10.229.12.212',
                'user': exasol_user,
                'password': exasol_password,
                'user_args': '-check-backup',
            }},
            'check_ssl_cert': {'args': '-w 60 -c 30 -H {host_ip} -n {domain_name} -r /etc/ssl/certs --altnames', 'parser': partial(self._to_dict, func=int),
                               'parameters': {'host_ip': '127.0.0.1', 'domain_name': 'www.example.com'}},
            'check-ldap-sync.pl': {'args': '', 'parser': json.loads},
        }

        self.__win_config = {
            'CheckCounter': {'args': '-a "Counter:ProcUsedMem=\\Process({process})\\Working Set" ShowAll MaxWarn=1073741824 MaxCrit=1073741824',
                             'parser': partial(self._to_dict_win, func=int), 'parameters': {'process': 'eo_server'}},
            'CheckCPU': {'args': '-a warn=100 crit=100 time=1 warn=100 crit=100 time=5 warn=100 crit=100 time=10',
                         'parser': partial(self._to_dict_win, func=int)},
            'CheckDriveSize': {'args': '-a CheckAll ShowAll perf-unit=M', 'parser': self._to_dict_win},
            'CheckEventLog': {'args': '-a file="{log}" MaxWarn=1 MaxCrit=1 "filter={query}" truncate=800 unique "syntax=%source% (%count%)"',
                              'parser': partial(self._to_dict_win, func=int), 'parameters': {'log': 'application',
                              'query': 'generated gt -7d AND type=\'error\''}},
            'CheckFiles': {'args': '-a "path={path}" "pattern={pattern}" "filter={query}" MaxCrit=1',
                           'parser': partial(self._to_dict_win, func=int),
                           'parameters': {'path': 'C:\\Import\\Exchange2Clearing', 'pattern': '*.*',
                           'query': 'creation lt -1h'}},
            'CheckLogFile': {'args': '-a file="{logfile}" column-split="{seperator}" "filter={query}"',
                             'parser': self._to_dict_win_log,
                             'parameters': {'logfile': 'c:\Temp\log\maxflow_portal.log', 'seperator': ' ',
                             'query': 'column4 = \'ERROR\' OR column4 = \'FATAL\''}},
            'CheckMEM': {'args': '-a MaxWarn=15G MaxCrit=15G ShowAll perf-unit=M type=physical type=page type=virtual',
                         'parser': self._to_dict_win},
            'CheckProcState': {'args': '-a ShowAll {process}', 'parser': self._to_dict_win_text,
                               'parameters': {'process': 'check_mk_agent.exe'}},
            'CheckServiceState': {'args': '-a ShowAll {service}', 'parser': self._to_dict_win_text,
                                  'parameters': {'service': 'ENAIO_server'}},
            'CheckUpTime': {'args': '-a MinWarn=1000d MinCrit=1000d', 'parser': partial(self._to_dict_win, func=int)},
        }

    def nrpe(self, check, timeout=60, **kwargs):
        config = self.__nrpe_config[check]
        parameters = {}
        parameters.update(config.get('parameters', {}))
        parameters.update(kwargs)
        pre_run_hook_ok = config.get('pre_run_hook', self._check_ok)
        if not pre_run_hook_ok(config.get('parameters', {})):
            raise CheckError('Pre run hook does not accept your parameters')
        cmd_args = config['args'].format(**parameters)
        cmd = shlex.split('/usr/lib/nagios/plugins/check_nrpe -u -H {h} -t {t} -c {c} {a}'.format(h=self.host,
                          t=timeout, c=check, a=cmd_args))
        try:
            output = subprocess32.check_output(cmd, shell=False, timeout=timeout)
        except subprocess32.CalledProcessError, e:
            e = fix_sub32_exc(e)
            if e.returncode < 3:
                output = str(e.output)
            else:
                output = str(e.output)
                return output
        except subprocess32.TimeoutExpired, e:
            e = fix_sub32_exc(e)
            raise NagiosError(str(e)), None, sys.exc_info()[2]
        logger.debug('output for cmd (%s): %s', cmd, output)
        return self.__nrpe_config[check]['parser'](output)

    def local(self, check, timeout=60, **kwargs):
        config = self.__local_config[check]
        parameters = {}
        parameters.update(config.get('parameters', {}))
        parameters.update(kwargs)
        pre_run_hook_ok = config.get('pre_run_hook', self._check_ok)
        if not pre_run_hook_ok(config.get('parameters', {})):
            raise CheckError('Pre run hook does not accept your parameters')
        cmd_args = config['args'].format(**parameters)
        cmd = shlex.split('/usr/lib/nagios/plugins/{c} {a}'.format(c=check, a=cmd_args))
        try:
            output = subprocess32.check_output(cmd, shell=False, timeout=timeout)
        except subprocess32.CalledProcessError, e:
            e = fix_sub32_exc(e)
            if e.returncode < 3:
                output = str(e.output)
            else:
                output = str(e.output)
                return output
        except subprocess32.TimeoutExpired, e:
            e = fix_sub32_exc(e)
            raise NagiosError(str(e)), None, sys.exc_info()[2]
        logger.debug('output for cmd (%s): %s', cmd, output)
        return self.__local_config[check]['parser'](output)

    def win(self, check, timeout=60, **kwargs):
        config = self.__win_config[check]
        parameters = {}
        parameters.update(config.get('parameters', {}))
        parameters.update(kwargs)
        pre_run_hook_ok = config.get('pre_run_hook', self._check_ok)
        if not pre_run_hook_ok(config.get('parameters', {})):
            raise CheckError('Pre run hook does not accept your parameters')
        cmd_args = config['args'].format(**parameters)
        cmd = shlex.split('/usr/lib/nagios/plugins/check_nrpe -u -H {h} -t {t} -c {c} {a}'.format(h=self.host,
                          t=timeout, c=check, a=cmd_args))
        try:
            output = subprocess32.check_output(cmd, shell=False, timeout=timeout)
        except subprocess32.CalledProcessError, e:
            e = fix_sub32_exc(e)
            if e.returncode < 3:
                output = str(e.output)
            else:
                output = str(e.output)
                return output
        except subprocess32.TimeoutExpired, e:
            e = fix_sub32_exc(e)
            raise NagiosError(str(e)), None, sys.exc_info()[2]
        logger.debug('output for cmd (%s): %s', cmd, output)
        return self.__win_config[check]['parser'](output)

    @staticmethod
    def _check_ok(config):
        '''Returns always True (ok) as result.'''

        return True

    @staticmethod
    def _check_path_chars(config):
        return re.match("^[a-zA-Z0-9\/_\-]+$", config['path'])

    @staticmethod
    @error_wrapped
    def _to_dict(output, func=float):
        return dict((a.split('=')[0], func(re.sub(r'.*?(-?[0-9]*\.[0-9]+|-?[0-9]+).*', r'\1', a.split('=')[1].split(';')[0]))) for a in
                    output.split('|')[-1].split())

    @staticmethod
    @error_wrapped
    def _to_dict_list(output):
        return dict((a, 1) for a in output.split('|')[-1].split())

    @staticmethod
    @error_wrapped
    def _to_dict_olderfiles(output):
        '''try to parse this output:
        OK - 0 files older as 600 min -- 0 files older as 540 min -- total 762 files -- older:

        >>> import json; json.dumps(NagiosWrapper._to_dict_olderfiles('OK - 0 files older as 600 min -- 112 files older as 60 min -- total 831 files -- older:'), sort_keys=True)
        '{"files older than time01": 112, "files older than time02": 0, "total files": 831}'
        '''

        return {'files older than time01': int(output.split(' -- ')[1].split()[0]), 'files older than time02': int(output.split(' -- ')[0].split()[2]), 'total files': int(output.split(' -- ')[2].split()[1])}

    @staticmethod
    @error_wrapped
    def _to_dict_win(output, func=float):
        '''try to parse this output:
        OK: physical memory: 4.8G, page file: 5.92G, virtual memory: 254M|'physical memory %'=29%;6;6 'physical memory'=5028644K;15728640;15728640;0;16776760 'page file %'=18%;53;53 'page file'=6206652K;15728640;15728640;0;33472700 'virtual memory %'=0%;99;99 'virtual memory'=259704K;15728640;15728640;0;8589934464
        '''

        return dict((a.split('=')[0], func(re.sub('[^0-9.]', '', a.split('=')[1].split(';')[0]))) for a in
                    shlex.split(output.split('|')[-1]))

    @staticmethod
    @error_wrapped
    def _to_dict_win_text(output):
        '''try to parse this output:
        OK: ENAIO_server: started
        '''

        po = {'status': output.split(':')[0], 'message': ' '.join(output.split(' ')[1:]).strip(' \n').split('|')[0]}
        if not po['status'] or not po['message']:
            raise NagiosError('Unable to parse {}'.format(output))
        return po

    @staticmethod
    @error_wrapped
    def _to_dict_win_log(output):
        '''try to parse this output:
        c:\Temp\log\maxflow_portal.log2014.04.29: 1 (2014-04-29 15:44:00,741 [5] WARN  BeckIT.MPO.... )
        '''

        if 'Nothing matched' in output:
            return {'count': 0}
        else:
            return {'count': int(output.split(' ')[1].strip(' '))}

    @staticmethod
    @error_wrapped
    def _to_dict_commitdiff(output):
        '''
        try to parse this output:
        CheckDiff OK - CommitLimit-Committed_AS: 24801200 | 24801200;1048576;524288
        '''

        return {(output.split(' ')[-4])[:-1]: int(output.split(' ')[-3])}

    @staticmethod
    @error_wrapped
    def _to_dict_logwatch(output):
        '''
        try to parse this output:
        WARNING - 0 new of 109 Lines on /access.log|newlines=0;100;5000;0;
        or
        OK - 0 new of 109 Lines on /access.log|newlines=0;1000;5000;0;
        '''

        return {'last': int(output.split(' ')[2]), 'total': int(output.split(' ')[5])}

    @staticmethod
    @error_wrapped
    def _to_dict_from_text(output):
        '''
        try to parse this output:
        Status ERROR : Avis-File for Fiege is missing
        or
        Status OK : file is ...
        '''

        return {'status': output.split(' ')[1].strip('\n'), 'message': ' '.join(output.split(' ')[3:]).strip('\n')}

    @staticmethod
    @error_wrapped
    def _to_dict_procs(output):
        '''
        try to parse this output:
        PROCS OK: 33 processes with command name 'httpd'
        '''

        return {'procs': int(output.split(' ')[2])}

    @staticmethod
    @error_wrapped
    def _to_dict_mysql_procs(output):
        '''
        try to parse this output:
        1 threads running. 0 seconds avg
        '''

        return {'threads': int(output.split(' ')[0]), 'avg': int(output.split(' ')[3])}

    @staticmethod
    @error_wrapped
    def _to_dict_mysql_slave(output):
        '''try to parse this output:
        Uptime: 38127526  Threads: 2  Questions: 42076974272  Slow queries: 1081545  Opens: 913119  Flush tables: 889  Open tables: 438  Queries per second avg: 1103.585 Slave IO: Yes Slave SQL: Yes Seconds Behind Master: 0
        '''

        po = dict(re.findall('([a-z][a-z0-9 ]+): ([a-z0-9.()]+)', output, re.IGNORECASE))
        for k, v in po.items():
            if not re.match('[a-z()]', v, re.IGNORECASE):
                po[k] = float(v)
        return po

    @staticmethod
    @error_wrapped
    def _to_dict_lounge_queries(output):
        '''try to parse this output:
        QUERY OK: 'SELECT COUNT(*) FROM global_orders WHERE ( billing_address LIKE '%Rollesbroich%' OR shipping_address LIKE '%Rollesbroich%' OR email LIKE '%Rollesbroich%' OR billing_address LIKE '%Süddeutsche TV%' OR shipping_address LIKE '%Süddeutsche TV%' OR email='sparfuchs-galileo@gmx.de' ) AND order_date >= DATE_SUB(CURDATE(),INTERVAL 1 DAY);' returned 0.000000
        '''

        return {'status': ' '.join(output.split()[:2]).strip(':'), 'query': ' '.join(output.split()[2:-2]),
                'result': float(output.split()[-1])}

    @staticmethod
    @error_wrapped
    def _to_dict_iostat(output):
        '''
        try to parse this output:
        OK - IOread 0.00 kB/s, IOwrite 214.80 kB/s  on sda with  31.10 tps |ioread=0.00;32000;64000;0;iowrite=214.80;30000;40000;0;
        '''

        return {'ioread': float(output.split(' ')[3]), 'iowrite': float(output.split(' ')[6]),
                'tps': float(output.split(' ')[13])}

    @staticmethod
    @error_wrapped
    def _to_dict_hpraid(output):
        '''
        try to parse this output:
        logicaldrive 1 (68.3 GB, RAID 1, OK) -- physicaldrive 1I:1:1 (port 1I:box 1:bay 1, SAS, 146 GB, OK) -- physicaldrive 1I:1:2 (port 1I:box 1:bay 2, SAS, 72 GB, OK) -- logicaldrive 2 (279.4 GB, RAID 1, OK) -- physicaldrive 1I:1:3 (port 1I:box 1:bay 3, SAS, 300 GB, OK) -- physicaldrive 1I:1:4 (port 1I:box 1:bay 4, SAS, 300 GB, OK) -- logicaldrive 3 (279.4 GB, RAID 1, OK) -- physicaldrive 2I:1:5 (port 2I:box 1:bay 5, SAS, 300 GB, OK) -- physicaldrive 2I:1:6 (port 2I:box 1:bay 6, SAS, 300 GB, OK) --
        '''

        return dict((b.split(' ')[0] + '_' + b.split(' ')[1], b.split(',')[-1].strip(' )')) for b in [a.strip(' --\n')
                    for a in output.split(' -- ')])

    @staticmethod
    @error_wrapped
    def _to_dict_newsletter(output):
        '''
        try to parse this output:
        OK: Not in timeframe (02:25:00 - 09:00:00)
        or
        CRITICAL: NL not processed for appdomain 17, Not processed for appdomain 18
        '''

        return {'status': output.split(': ')[0], 'messages': output.split(': ')[-1].strip('\n').split(',')}

    @staticmethod
    @error_wrapped
    def _to_dict_hpasm(output):
        '''
        try to parse this output:
        OK - System: 'proliant dl360 g6', S/N: 'CZJ947016M', ROM: 'P64 05/05/2011', hardware working fine, da: 3 logical drives, 6 physical drives cpu_0=ok cpu_1=ok ps_2=ok fan_1=46% fan_2=46% fan_3=46% fan_4=46% temp_1=21 temp_2=40 temp_3=40 temp_4=35 temp_5=34 temp_6=37 temp_7=32 temp_8=36 temp_9=32 temp_10=36 temp_11=32 temp_12=33 temp_13=48 temp_14=29 temp_15=32 temp_16=30 temp_17=29 temp_18=39 temp_19=37 temp_20=38 temp_21=45 temp_22=42 temp_23=39 temp_24=48 temp_25=35 temp_26=46 temp_27=35 temp_28=71 | fan_1=46%;0;0 fan_2=46%;0;0 fan_3=46%;0;0 fan_4=46%;0;0 'temp_1_ambient'=21;42;42 'temp_2_cpu#1'=40;82;82 'temp_3_cpu#2'=40;82;82 'temp_4_memory_bd'=35;87;87 'temp_5_memory_bd'=34;78;78 'temp_6_memory_bd'=37;87;87 'temp_7_memory_bd'=32;78;78 'temp_8_memory_bd'=36;87;87 'temp_9_memory_bd'=32;78;78 'temp_10_memory_bd'=36;87;87 'temp_11_memory_bd'=32;78;78 'temp_12_power_supply_bay'=33;59;59 'temp_13_power_supply_bay'=48;73;73 'temp_14_memory_bd'=29;60;60 'temp_15_processor_zone'=32;60;60 'temp_16_processor_zone'=3
        or
        OK - ignoring 16 dimms with status 'n/a' , System: 'proliant dl360p gen8', S/N: 'CZJ2340R6C', ROM: 'P71 08/20/2012', hardware working fine, da: 1 logical drives, 4 physical drives
        '''

        return {'status': output.split(' - ')[0], 'message': output.split(' - ')[-1].strip('\n')}

    @staticmethod
    @error_wrapped
    def _parse_memory(output):
        result = dict((a.split('=')[0], a.split('=')[1].split(';')[0]) for a in output.split('|')[-1].split())
        conv = {
            't': 1000000000,
            'tb': 1000000000,
            'tib': 1073741824,
            'g': 1000000,
            'gb': 1000000,
            'gib': 1048576,
            'm': 1000,
            'mb': 1000,
            'mib': 1024,
            'k': 1,
            'kb': 1,
            'kib': 1.024,
        }
        p = re.compile(r'^(\d+(?:\.\d+)?)(?:\s+)?(\w+)?$', re.I)

        for k in result:
            parts = p.findall(result[k])
            if len(parts):
                v, u = parts[0]
                result[k] = (float(v) * conv[u.lower()] if u and u.lower() in conv else float(v))

        return result

    @staticmethod
    @error_wrapped
    def _parse_disks(output):
        '''try to parse output of /usr/lib/nagios/plugins/check_disk -w 10% -c 5%  -u MB

        >>> import json; json.dumps(NagiosWrapper._parse_disks('DISK OK - free space: / 80879 MB (69% inode=99%); /dev 64452 MB (99% inode=99%); /selinux 0 MB (100% inode=99%); | /=35303MB;110160;116280;0;122401 /dev=0MB;58006;61229;0;64452 /selinux=0MB;0;0;0;0'), sort_keys=True)
        '{"/": {"total_mb": 116280, "used_mb": 35303}, "/dev": {"total_mb": 61229, "used_mb": 0}}'
        '''

        r = re.compile('[^0-9.]')
        performance_data = output.split('|')[-1]
        result = dict((m[0], {'used_mb': int(r.sub('', f[0])), 'total_mb': int(int(f[4]) * USABLE_DISKSPACE_FACTOR)})
                      for (m, f) in [[s.split(';') for s in a.split('=')] for a in performance_data.split()]
                      if int(f[4]) != 0)
        return result


if __name__ == '__main__':
    if len(sys.argv) == 3:
        checkname = sys.argv[2]
        check = NagiosWrapper(sys.argv[1])
        print json.dumps(check.nrpe(checkname), indent=4)
    elif len(sys.argv) == 4:
        checkname = sys.argv[2]
        query = sys.argv[3]
        check = NagiosWrapper(sys.argv[1])
        print json.dumps(check.win(checkname, query=query), indent=4)
    elif len(sys.argv) == 5:
        checkname = sys.argv[2]
        path = sys.argv[3]
        query = sys.argv[4]
        check = NagiosWrapper(sys.argv[1])
        print json.dumps(check.win(checkname, path=path, query=query), indent=4)
    elif len(sys.argv) == 6:
        checkname = sys.argv[2]
        directory = sys.argv[3]
        epoch = sys.argv[4]
        name = sys.argv[5]
        check = NagiosWrapper(sys.argv[1])
        print json.dumps(check.nrpe(checkname, directory=directory, epoch=epoch, name=name), indent=4)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import jinja2
import os
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from notification import BaseNotification

import logging
logger = logging.getLogger(__name__)

thisdir = os.path.join(os.path.dirname(__file__))
template_dir = os.path.join(thisdir, '../templates/mail')
jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_dir))


class Mail(BaseNotification):

    @classmethod
    def send(cls, alert, *args, **kwargs):
        sender = cls._config.get('notifications.mail.sender')
        subject = cls._get_subject(alert, custom_message=kwargs.get('subject'))
        html = kwargs.get('html', False)
        cc = kwargs.get('cc', [])
        hide_recipients = kwargs.get('hide_recipients', True)
        repeat = kwargs.get('repeat', 0)
        expanded_alert_name = cls._get_expanded_alert_name(alert)

        try:
            tmpl = jinja_env.get_template('alert.txt')
            body_plain = tmpl.render(expanded_alert_name=expanded_alert_name, **alert)
        except Exception:
            logger.exception('Error parsing email template for alert %s with id %s', alert['name'], alert['id'])
        else:
            if html:
                msg = MIMEMultipart('alternative')
                tmpl = jinja_env.get_template('alert.html')
                body_html = tmpl.render(expanded_alert_name=expanded_alert_name, **alert)
                part1 = MIMEText(body_plain.encode('utf-8'), 'plain', 'utf-8')
                part2 = MIMEText(body_html.encode('utf-8'), 'html', 'utf-8')
                msg.attach(part1)
                msg.attach(part2)
            else:
                msg = MIMEText(body_plain.encode('utf-8'), 'plain', 'utf-8')

            msg['Subject'] = subject
            msg['From'] = 'ZMON 2 <{}>'.format(sender)

            args = BaseNotification.resolve_group(args)

            if hide_recipients:
                msg['To'] = 'Undisclosed Recipients <{}>'.format(sender)
                msg['Bcc'] = ', '.join(args)
            else:
                msg['To'] = ', '.join(args)
            msg['Cc'] = ', '.join(cc)

            mail_host = cls._config.get('notifications.mail.host', 'localhost')
            mail_port = cls._config.get('notifications.mail.port', '25')

            logger.info("Relaying via %s %s", mail_host, mail_port)

            if cls._config.get('notifications.mail.on', True):
                try:
                    if mail_host != 'localhost':
                        s = smtplib.SMTP_SSL(mail_host, mail_port)
                    else:
                        s = smtplib.SMTP(mail_host, mail_port)

                except Exception:
                    logger.exception('Error connecting to SMTP server %s for alert %s with id %s',
                                     mail_host, alert['name'], alert['id'])
                else:
                    try:
                        mail_user = cls._config.get('notifications.mail.user', None)
                        if mail_user is not None:
                            s.login(mail_user, cls._config.get('notifications.mail.password'))

                        s.sendmail(sender, list(args) + cc, msg.as_string())
                    except SMTPAuthenticationError:
                        logger.exception('Error sending email for alert %s with id %s: authentication failed for %s', alert['name'], alert['id'], mail_user)
                    except Exception:
                        logger.exception('Error sending email for alert %s with id %s', alert['name'], alert['id'])
                    finally:
                        s.quit()
        finally:
            return repeat


if __name__ == '__main__':
    import sys
    Mail.send({'entity': {'id': 'test'}, 'value': 5}, *sys.argv[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

#from cassandra.io.libevreactor import LibevConnection
from cassandra.cluster import Cluster

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial

logger = logging.getLogger('zmon-worker.cassandra-function')

class CassandraFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(CassandraFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(CassandraWrapper, node=factory_ctx.get('host'))

class CassandraWrapper(object):

    def __init__(self, node, keyspace, port=9042, connect_timeout=1):
        # for now using a single host / node should be seed nodes or at least available nodes
        self.node = node
        self.port = port
        self.keyspace = keyspace
        self.connect_timeout = connect_timeout

    def execute(self, stmt):
        cl = Cluster([self.node], connect_timeout=self.connect_timeout)
        #cl.connection_class = LibevConnection

        session = None
        try:
            session = cl.connect()
            session.set_keyspace(self.keyspace)

            rs = session.execute(stmt)

            result = []

            for r in rs:
                result.append(r)

            return result

        finally:
            cl.shutdown();

        return {}


# -*- coding: utf-8 -*-

"""
Project settings:
 To customize the settings for a local environment please create another module called settings_local.py and change
 there the values you want.

You don't have to duplicate all the logic in settings_dev.py inside settings_local.py, instead you import it and
can modify certain values from settings_dev. For example to modify only the log level:

##### content of file: settings_local.py

from settings_dev import *
LOGGING['loggers']['']['level'] = 'INFO'
RPC_SERVER_CONF['LOGGING']['loggers']['']['level'] = 'INFO'

#### END of settings_local.py
"""


EXTERNAL_CONFIG = {}


try:
    from settings_local import *
except ImportError:
    # no settings_local.py found, falling back to settings_pro.py
    from settings_pro import *


def set_workers_log_level(level):
    if level:
        LOGGING['loggers']['']['level'] = level


def set_rpc_server_port(port):
    if port and str(port).isdigit():
        RPC_SERVER_CONF['PORT'] = int(port)


def set_external_config(external_config):
    global EXTERNAL_CONFIG
    EXTERNAL_CONFIG = dict(external_config)


def get_external_config():
    return EXTERNAL_CONFIG



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from pymongo import MongoClient

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial

logger = logging.getLogger('zmon-worker.mongodb-function')

class MongoDBFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(MongoDBFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(MongoDBWrapper, host=factory_ctx.get('host'))

class MongoDBWrapper(object):

    def __init__(self, host, port=27017):
        self.host = host
        self.port = port

    def find(self, database, collection, query, limit=50):
        client = MongoClient(self.host, self.port)
        try:
            db = client[database]
            rs = db[collection].find(query, limit=limit)
            result = []

            for r in rs:
                result.append(r)

            return result

        finally:
            client.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module is intended as a quick and dirty drop in replacement of kombu module, covering only
the functionality we use in zmon. In essence we are trying to cover these use cases:
    from kombu.connection import Connection
    try:
        c = Connection(config['backend'])
    except KeyError:
        raise Exception('"backend" property missing in config object')

    r_conn = redis.StrictRedis(host=c.hostname, port=c.port, db=c.virtual_host)
And:
    from kombu import Queue
    app.conf.update(CELERY_QUEUES=(Queue('zmon:queue:default',routing_key='default'),
                    Queue('zmon:queue:secure', routing_key='secure'), Queue('zmon:queue:snmp', routing_key='snmp')))
"""

import re
from collections import namedtuple


class Queue(object):

    '''
    Used to emulate kombu Queue
    '''

    def __init__(self, queue, routing_key=None):
        self.queue = queue
        self.routing_key = routing_key


def parse_redis_conn(conn_str):
    '''
    Emulates kombu.connection.Connection that we were using only to parse redis connection string
    :param conn_str: example 'redis://localhost:6379/0'
    :return: namedtuple(hostname, port, virtual_host)
    '''

    Connection = namedtuple('Connection', 'hostname port virtual_host')
    conn_regex = r'redis://([-.a-zA-Z0-9_]+):([0-9]+)/([0-9]+)'
    m = re.match(conn_regex, conn_str)
    if not m:
        raise Exception('unable to parse redis connection string: {}'.format(conn_str))
    return Connection(m.group(1), int(m.group(2)), m.group(3))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import requests
from notification import BaseNotification
import zmon_worker_monitor.eventloghttp as eventlog
import logging

logger = logging.getLogger(__name__)


SMS_PROVIDER_URL = 'https://gateway.smstrade.de'
SMS_SENDER = 'zmon2'
SMS_API_KEY = ''
SMS_ROUTE = 'gold'
SMS_MAXLENGTH = 2048

#logger = get_task_logger('zmon-worker')


class SmsException(Exception):

    pass


class Sms(BaseNotification):

    @classmethod
    def send(cls, alert, *args, **kwargs):
        provider_url = cls._config.get('notifications.sms.provider_url', SMS_PROVIDER_URL)
        phone_numbers = BaseNotification.resolve_group(args, phone=True)
        repeat = kwargs.get('repeat', 0)

        maxlen = cls._config.get('notifications.sms.maxlength', SMS_MAXLENGTH)
        message = cls._get_subject(alert, custom_message=kwargs.get('message'))[:maxlen]

        request_params = {
            'to': '',
            'key': cls._config['notifications.sms.apikey'],
            'from': cls._config.get('notifications.sms.sender', SMS_SENDER),
            'route': cls._config.get('notifications.sms.route', SMS_ROUTE),
            'message': message,
            'cost': 1,
            'message_id': 1,
        }

        alert_id = alert.get('alert_def', {}).get('id', 0)
        entity = alert.get('entity', {}).get('id', 0)

        try:
            if cls._config.get('notifications.sms.on', True):
                for phone in phone_numbers:
                    request_params['to'] = phone
                    r = requests.get(provider_url, params=request_params, verify=False)
                    url_secured = r.url.replace(request_params['key'], '*' * len(request_params['key']))
                    logger.info('SMS sent: request to %s --> status: %s, response headers: %s, response body: %s',
                                url_secured, r.status_code, r.headers, r.text)
                    r.raise_for_status()
                    eventlog.log(cls._EVENTS['SMS_SENT'].id, alertId=alert_id, entity=entity, phoneNumber=phone,
                                 httpStatus=r.status_code)
        except Exception:
            logger.exception('Failed to send sms for alert %s with id %s to: %s', alert['name'], alert['id'],
                             list(phone_numbers))
        finally:
            return repeat


if __name__ == '__main__':

    Sms.update_config({
        'notifications.sms.on': True,
        'notifications.sms.apikey': '--secret--',
        'notifications.sms.sender': 'zmon2',
        'notifications.sms.route': 'gold',
    })

    test_recipients = ['1stlevel']

    fake_alert = {
        'is_alert': True,
        'alert_def': {'name': 'Test'},
        'entity': {'id': 'hostxy'},
        'captures': {},
    }

    Sms.send(fake_alert, *test_recipients, **{'message': 'My customized zmon2 alert message'})
    Sms.send(fake_alert, *test_recipients)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''General math/aggregate functions
'''

from functools import partial
import functional
import math


def _percentile(N, percent, key=functional.id):
    """
    Find the percentile of a list of values.

    @parameter N - is a list of values. Note N MUST BE already sorted.
    @parameter percent - a float value from 0.0 to 1.0.
    @parameter key - optional key function to compute value from each element of N.

    @return - the percentile of the values

    >>> percentile([0,1], 0.5)
    0.5

    >>> percentile([0,1,2], 0.9)
    1.8

    >>> percentile([], 0.9) is None
    True
    """

    if not N:
        return None
    k = (len(N) - 1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return key(N[int(k)])
    d0 = key(N[int(f)]) * (c - k)
    d1 = key(N[int(c)]) * (k - f)
    return d0 + d1


# median is 50th percentile.
_median = partial(_percentile, percent=0.5)


def median(results):
    return _median(sorted(results))


def percentile(results, percent):
    return _percentile(sorted(results), percent)


def apply_aggregate_function(results, func, key=functional.id, **args):
    '''
    >>> apply_aggregate_function([1,2,3], sum)
    6
    >>> apply_aggregate_function([{'a': 0}, {'a': 2}], _percentile, key=lambda x:x['a'], percent=0.9)
    1.8
    '''

    return func(map(key, results), **args)


def delta(results):
    '''
    >>> delta([])
    0

    >>> delta([0, 10])
    10

    >>> delta([10, 0])
    -10
    '''

    if not results:
        # no results => zero delta
        return 0
    return results[-1] - results[0]


def avg(results):
    '''
    >>> avg([]) is None
    True

    >>> avg([0, 1])
    0.5
    '''

    if not results:
        return None
    return sum(results) * 1.0 / len(results)


def first(results):
    '''
    >>> first([1, 2])
    1

    >>> first([]) is None
    True
    '''

    return (results[0] if results else None)


def _min(results):
    '''
    >>> _min([2, 1, 3, 2])
    1

    >>> _min([]) is None
    True
    '''

    return (min(results) if results else None)


def _max(results):
    '''
    >>> _max([2, 1, 3, 2])
    3

    >>> _max([]) is None
    True
    '''

    return (max(results) if results else None)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import ldap
try:
    import ldapapi
except:
    ldapapi = None
import logging
import time

from zmon_worker_monitor.zmon_worker.errors import CheckError
from ldap.dn import explode_dn

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial
from zmon_worker_monitor import plugin_manager


STATISTICS_OPERATIONS_TO_MONITOR = frozenset([
    'bind',
    'unbind',
    'search',
    'modify',
    'add',
    'delete',
    'extended',
])

STATISTICS_GAUGE_KEYS = frozenset([
    'threads_active',
    'threads_max',
    'connections_current',
    'statistics_entries',
    'waiters_read',
    'waiters_write',
    'connections_max_file_descriptors',
])

# rename some keys to make the names more friendly
STATISTICS_FRIENDLY_KEY_NAMES = {'statistics_entries': 'entries',
                                 'connections_max_file_descriptors': 'max_file_descriptors'}


class LdapFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(LdapFactory, self).__init__()
        # fields from config
        self._ldapuser = None
        self._ldappass = None
        # fields to store dependencies: plugin depends on 1 other plugins
        self.counter_factory = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self._ldapuser = conf['user']
        self._ldappass = conf['pass']

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        if not self.counter_factory:
            self.counter_factory = plugin_manager.get_plugin_obj_by_name('counter', 'Function')

        return propartial(LdapWrapper, user=self._ldapuser, password=self._ldappass, host=factory_ctx['host'],
                          counter=self.counter_factory.create(factory_ctx))


class UnsupportedMethodException(Exception):

    pass


class DuplicateBindException(Exception):

    pass


class LdapWrapper(object):

    def __init__(
        self,
        host=None,
        tls=True,
        krb5=False,
        user='uid=nagios,ou=robots,ou=users,dc=example,dc=com',
        password='',
        timeout=60,
        logger=None,
        counter=None,
    ):
        '''expects ready to use "partial" for counter (CounterWrapper)'''

        self.logger = logger or logging.getLogger()
        self._counter = counter('')
        self.use_tls = tls
        self.use_krb5 = krb5
        self.user_base_dn = ','.join(explode_dn(user)[1:])
        self.user_filter = '(' + explode_dn(user)[0] + ')'
        self.user_attrib = ['dn']
        # password auth
        self.bind_dn = user
        self.__password = password
        self.session = None
        self.host = host
        if self.host and len(self.host.split('.')) <= 1:
            # FIXME
            self.host += '.zalando'

    def _connect(self, ldapserver):
        if self.session:
            raise CheckError('LDAP Error: duplicate bind exception.')
        self.logger.debug('connect to %s', ldapserver)
        uri = ldapserver
        if not uri.startswith("ldap://") and not uri.startswith("ldaps://"):
            uri = 'ldap://{0}'.format(ldapserver)
        try:
            if self.use_krb5 and self.use_tls:
                self.logger.debug('sasl bind')
                self.session = ldapapi.Session(uri, tls=True)
            elif self.use_tls:
                self.logger.debug('simple bind')
                self.session = ldapapi.Session(uri, self.bind_dn, self.__password, tls=True)
            elif uri.startswith("ldaps://"):
                self.logger.debug('LDAPS + simple bind')
                self.session = ldapapi.Session(uri, self.bind_dn, self.__password, tls=False)
            else:
                raise CheckError('LDAP Error: unsupported method exception.')
        except CheckError:
            raise
        except Exception, e:
            raise CheckError('Error connecting to LDAP: {}'.format(e)), None, sys.exc_info()[2]

    def _search(self, base, fltr, attrs, scope=ldap.SCOPE_SUBTREE):
        return self.session.search(base, fltr, attrs, scope)

    def _disconnect(self):
        self.logger.debug('disconnect')
        try:
            self.session.disconnect()
        except Exception, e:
            raise CheckError('Error disconnecting to LDAP: {}'.format(e)), None, sys.exc_info()[2]
        self.session = None

    def _sync_state(self, ldapserver):
        base = 'dc=example,dc=com'
        fltr = '(objectclass=*)'
        attr = ['contextCSN']
        self._connect(ldapserver)
        result = self._search(base, fltr, attr, scope=ldap.SCOPE_BASE)
        self._disconnect()
        return result[0][1]['contextCSN']

    def _get_rid_to_url(self, ldapserver):
        '''Returns a dict for this query:
        % ldapsearch -b cn=config '(cn=config)' -H ldap://myserver olcServerID
        '''

        rid2url = {}
        url2rid = {}
        base = 'cn=config'
        fltr = '(cn=config)'
        attr = ['olcServerID']
        self._connect(ldapserver)
        res = self.session.conn.search_ext_s(base, ldap.SCOPE_BASE, fltr, attr)
        self._disconnect()
        for rid_url in res[0][1]['olcServerID']:
            rid, url = rid_url.split()
            rid = '%03d' % int(rid)
            rid2url[rid] = url
            url2rid[url] = rid
        return rid2url, url2rid

    def _get_timestamp_rid(self, csn):
        '''csn: 20140227093429.363252Z#000000#004#000000'''

        ts, _, rid, _ = csn.split('#')
        ts = ts.split('.')[0]
        # ts = datetime.datetime.strptime(ts, "%Y%m%d%H%M%S")
        ts = int(ts)
        return rid, ts

    def sync(self):
        '''Example:
        checkldap().sync() => [{'newest': 20140516151002, 'elapsed': 0.14442706108093262, 'ok': True, 'diff': 0, 'server': 'myserv'}, {'newest': 20140516151002, 'elapsed': 0.19423580169677734, 'ok': True, 'diff': 0, 'server': 'myserver'}, {'newest': 20140516151002, 'elapsed': 0.2617530822753906, 'ok': True, 'diff': 0, 'server': 'z-auth123.example'}, {'newest': 20140516151002, 'elapsed': 0.15635299682617188, 'ok': True, 'diff': 0, 'server': 'myserver'}, {'newest': 20140516151002, 'elapsed': 0.20283913612365723, 'ok': True, 'diff': 0, 'server': 'myserver'}]
        '''

        try:
            rid2url, url2rid = self._get_rid_to_url(self.host)
            ldapservers = map(lambda url: url[7:], url2rid.keys())
            return self._sync(ldapservers)
        except CheckError:
            raise
        except Exception, e:
            raise CheckError('{}'.format(e)), None, sys.exc_info()[2]

    def _sync(self, ldapservers):
        '''Returns a list of dict, where 'diff' is the difference to the 'newest' of the full list, 'newest' is the newest timestamp for the given 'server', 'ok' means LDAP state for current 'server' and 'elapsed' the runtime of that ldap request.
        Example dict:
        {'diff': 0,
        'elapsed': 0.2969970703125,
        'newest': 20140228135148,
        'ok': True,
        'server': 'myserver'}
        '''

        if not ldapservers:
            return
        results = []
        for ldapserver in ldapservers:
            try:
                start = time.time()
                csn_list = self._sync_state(ldapserver)
                rid2ts = {}
                rid_ts = map(lambda csn: self._get_timestamp_rid(csn), csn_list)
                newest = rid_ts[0][1]
                for rid, ts in rid_ts:
                    rid2ts[rid] = ts
                    if ts > newest:
                        newest = ts
                elapsed = time.time() - start
                results.append({
                    'server': ldapserver,
                    'ok': True,
                    'newest': newest,
                    'elapsed': elapsed,
                })
            except ldap.LOCAL_ERROR:
                bind_type = 'simple bind'
                if self.use_krb5:
                    bind_type = 'sasl bind'
                msg = 'Could not connect to {} via {}'.format(ldapserver, bind_type)
                self.logger.exception(msg)
                raise CheckError(msg)
            except ldap.CONFIDENTIALITY_REQUIRED:
                results.append({'ok': False, 'server': ldapserver})
        newest = 0
        for result in results:
            if result['ok']:
                if result['newest'] > newest:
                    newest = result['newest']
        for result in results:
            if result['ok']:
                result['diff'] = newest - result['newest']
        return results

    def auth(self):
        try:
            start = time.time()
            self._connect(self.host)
            connect_elapsed = time.time() - start
            self._search(self.user_base_dn, self.user_filter, self.user_attrib)
            all_elapsed = time.time() - start
            search_elapsed = all_elapsed - connect_elapsed
            self._disconnect()
            return {
                'ok': True,
                'connect_time': connect_elapsed,
                'search_time': search_elapsed,
                'elapsed': all_elapsed,
            }
        except ldap.LOCAL_ERROR:
            bind_type = 'simple bind'
            if self.use_krb5:
                bind_type = 'sasl bind'
            msg = 'Could not connect to {} via {}'.format(self.host, bind_type)
            self.logger.exception(msg)
            raise CheckError(msg)
        except ldap.CONFIDENTIALITY_REQUIRED:
            return {'ok': False}
        except Exception, e:
            raise CheckError('Error authenticating to LDAP: {}'.format(e)), None, sys.exc_info()[2]

    @staticmethod
    def _split_monitor_dn(dn):
        '''
        >>> LdapWrapper._split_monitor_dn('cn=Max File Descriptors,cn=Connections,cn=Monitor')
        ('connections', 'max_file_descriptors')
        '''

        parts = dn.replace(' ', '_').split(',')
        return (parts[1])[3:].lower(), (parts[0])[3:].lower()

    def statistics_raw(self):
        '''collect statistics from OpenLDAP "Monitor" DB as a dict

        ldapsearch -b cn=Connections,cn=Monitor -H ldap://myserver '(monitorCounter=*)' '+'

        Example result::

            {
              "connections_current": "51",
              "connections_max_file_descriptors": "65536",
              "connections_total": "26291",
              "operations_add_completed": "0",
              "operations_add_initiated": "0",
              "operations_bind_completed": "25423",
              "operations_bind_initiated": "25423",
              "operations_delete_completed": "0",
              "operations_delete_initiated": "0",
              "operations_extended_completed": "293",
              "operations_extended_initiated": "293",
              "operations_modify_completed": "91",
              "operations_modify_initiated": "91",
              "operations_search_completed": "22865",
              "operations_search_initiated": "22866",
              "operations_unbind_completed": "25233",
              "operations_unbind_initiated": "25233",
              "statistics_bytes": "122581936",
              "statistics_entries": "64039",
              "statistics_pdu": "112707",
              "statistics_referrals": "0",
              "waiters_read": "51",
              "waiters_write": "0"
            }
        '''

        try:
            self._connect(self.host)
            data = {}
            # we need to use the internal "conn" attribute as the default _search is using paging which does not work for the "cn=Monitor" tree!
            result = self.session.conn.search_s('cn=Monitor', ldap.SCOPE_SUBTREE, '(objectClass=monitorCounterObject)',
                                                ['monitorCounter'])
            for dn, attr in result:
                category, counter = self._split_monitor_dn(dn)
                data['{}_{}'.format(category, counter)] = int(attr['monitorCounter'][0])

            result = self.session.conn.search_s('cn=Threads,cn=Monitor', ldap.SCOPE_SUBTREE,
                                                '(&(objectClass=monitoredObject)(monitoredInfo=*))', ['monitoredInfo'])
            for dn, attr in result:
                category, key = self._split_monitor_dn(dn)
                if key in ('active', 'max'):
                    data['{}_{}'.format(category, key)] = int(attr['monitoredInfo'][0])

            result = self.session.conn.search_s('cn=Operations,cn=Monitor', ldap.SCOPE_SUBTREE,
                                                '(objectClass=monitorOperation)', ['monitorOpInitiated',
                                                'monitorOpCompleted'])
            for dn, attr in result:
                category, op = self._split_monitor_dn(dn)
                if op in STATISTICS_OPERATIONS_TO_MONITOR:
                    data['{}_{}_initiated'.format(category, op)] = int(attr['monitorOpInitiated'][0])
                    data['{}_{}_completed'.format(category, op)] = int(attr['monitorOpCompleted'][0])
            self._disconnect()
        except CheckError:
            raise
        except Exception, e:
            raise CheckError('{}'.format(e)), None, sys.exc_info()[2]
        return data

    def statistics(self):
        '''uses raw statistics and computes counter values (i.e. e.g. operations per second)

        Example result::

            {
              "connections_current": 74,
              "connections_per_sec": 24.1,
              "entries": 353540,
              "max_file_descriptors": 65536,
              "operations_add_per_sec": 0.0,
              "operations_bind_per_sec": 26.4,
              "operations_delete_per_sec": 0.0,
              "operations_extended_per_sec": 1.15,
              "operations_modify_per_sec": 0.0,
              "operations_search_per_sec": 20.66,
              "operations_unbind_per_sec": 24.1,
              "threads_active": 2,
              "threads_max": 16,
              "waiters_read": 73,
              "waiters_write": 0
            }
        '''

        _data = self.statistics_raw()
        data = {}
        for key, val in _data.items():
            if key in STATISTICS_GAUGE_KEYS:
                data[STATISTICS_FRIENDLY_KEY_NAMES.get(key, key)] = val
            elif key == 'connections_total':
                data['connections_per_sec'] = round(self._counter.key(key).per_second(val), 2)
            elif key.startswith('operations_') and key.endswith('_initiated'):
                data[key.replace('_initiated', '_per_sec')] = round(self._counter.key(key).per_second(val), 2)

        # gc_count = round(self._counter.key('gcCount').per_second(gc_count), 2)
        return data






from notification import BaseNotification

import logging
import requests
import urllib
import json
logger = logging.getLogger(__name__)


class NotifyHipchat(BaseNotification):

    @classmethod
    def send(cls, alert, *args, **kwargs):
        url = cls._config.get('notifications.hipchat.url')
        token = kwargs.get('token', cls._config.get('notifications.hipchat.token'))
        repeat = kwargs.get('repeat', 0)

        color = 'green' if alert and not alert.get('is_alert') else kwargs.get("color", "red")

        message = {"message": kwargs.get("message", cls._get_subject(alert)), "color": color}

        try:
            logger.info("Sending to: " + '{}/v2/room/{}/notification?auth_token={}'.format(url, urllib.quote(kwargs['room']), token) + " " + json.dumps(message))
            r = requests.post('{}/v2/room/{}/notification?auth_token={}'.format(url, urllib.quote(kwargs['room']), token), data=json.dumps(message), verify=False, headers={'Content-type':'application/json'})
            r.raise_for_status()
        except Exception as ex:
            logger.exception("Hipchat write failed %s", ex)

        return repeat


# -*- coding: utf-8 -*-
"""
Project settings for development:
 To customize the settings for a local environment please create another module called settings_local.py and change
 there the values you want, they will override the ones in this file
"""

import os

app_home = os.path.abspath(os.environ['APP_HOME'] if 'APP_HOME' in os.environ else './')

data_dir = os.path.abspath(os.path.join(app_home, 'zmon_worker_data'))

# application data folder needs to be created by the application itself
for d in (data_dir, ):
    if not os.path.isdir(d):
        os.mkdir(d)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(module)s - %(process)d - %(thread)d - %(message)s'
        },
        'custom': {
            'format': '%(asctime)s - %(levelname)s - %(name)s - %(funcName)s - %(message)s'
        },
    },

    'handlers': {

        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'custom'
        },

    },

    'loggers': {
        '': {
            'handlers': ['console'],
            'propagate': True,
            'level': 'DEBUG',
        },
    }
}


RPC_SERVER_CONF = dict(

    HOST='localhost',

    PORT=8500,

    RPC_PATH='/zmon_rpc',

    LOGGING={
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'verbose': {
                'format': '%(asctime)s - %(name)s - %(levelname)s - %(module)s - %(process)d - %(thread)d - %(message)s'
            },
            'custom': {
                'format': '%(asctime)s - %(levelname)s - %(name)s - %(funcName)s - %(message)s'
            },
        },
        'handlers': {

            'console': {
                'level': 'INFO',
                'class': 'logging.StreamHandler',
                'formatter': 'custom'
            },

        },
        'loggers': {
            '': {
                'handlers': ['console'],
                'propagate': True,
                'level': 'DEBUG',
            },
        }
    }
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod


class IBasePlugin(object):
    """
    Base class for all adapters (plugin types). Users should not extend this class directly.
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        """
        Set the basic variables.
        """
        self.is_activated = False

    def activate(self):
        """
        Called at plugin activation.
        """
        self.is_activated = True

    def deactivate(self):
        """
        Called when the plugin is disabled.
        """
        self.is_activated = False

    @abstractmethod
    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        raise NotImplementedError




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from ibase_plugin import IBasePlugin
from abc import ABCMeta, abstractmethod


class IFunctionFactoryPlugin(IBasePlugin):
    """
    Base class for all plugins of type Function. We call it Function adapter.
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        super(IFunctionFactoryPlugin, self).__init__()

    @abstractmethod
    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: creation context
        :return: an object that implements a check function
        """
        raise NotImplementedError


class ProtectedPartial(object):

    '''
    Provides functools.partial functionality with one additional feature: if keyword arguments contain '__protected'
    key with list of arguments as value, the appropriate values will not be overwritten when calling the partial. This
    way we can prevent user from overwriting internal zmon parameters in check command. The protected key uses double
    underscore to prevent overwriting it, we reject all commands containing double underscores.
    '''

    def __init__(self, func, *args, **kwargs):
        self.__func = func
        self.__partial_args = args
        self.__partial_kwargs = kwargs
        self.__protected = frozenset(kwargs.get('__protected', []))
        self.__partial_kwargs.pop('__protected', None)

    def __call__(self, *args, **kwargs):
        new_kwargs = self.__partial_kwargs.copy()
        new_kwargs.update((k, v) for (k, v) in kwargs.iteritems() if k not in self.__protected)
        return self.__func(*self.__partial_args + args, **new_kwargs)


def propartial(func, *args, **kwargs):
    '''
    >>> propartial(int, base=2)('100')
    4
    >>> propartial(int, base=2)('100', base=16)
    256
    >>> propartial(int, base=2, __protected=['base'])('100', base=16)
    4
    '''

    return ProtectedPartial(func, *args, **kwargs)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import re
from zmon_worker_monitor.zmon_worker.errors import DbError

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


DEFAULT_PORT = 3306
MAX_RESULTS = 100

CONNECTION_RE = \
    re.compile(r'''
^(?P<host>[^:/]+)       # host - either IP o hostname
(:(?P<port>\d+))?       # port - integer, optional
/(?P<dbname>\w+)        # database name
$
'''
               , re.X)


class MySqlFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(MySqlFactory, self).__init__()
        # fields from config
        self._user = None
        self._pass = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self._user = conf['user']
        self._pass = conf['pass']

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(MySqlWrapper, shards=factory_ctx['shards'], user=self._user, password=self._pass,
                          timeout=factory_ctx['soft_time_limit'] * 1000)


def _import_db_driver():
    module_alternatives = 'MySQLdb', 'pymysql'
    for module in module_alternatives:
        try:
            return __import__(module, globals(), locals(), [], -1)
        except Exception, e:
            if module == module_alternatives[-1]:
                raise
            else:
                # print is well supported by celery, this will end up as a log entry
                print 'Warning: Import of module {} failed: {}'.format(module, e)


class MySqlWrapper(object):

    '''Shard-aware SQL adapter
    sql().execute('SELECT 1').result()
    '''

    def __init__(self, shards, user='nagios', password='', timeout=60000, shard=None):
        '''
        Parameters
        ----------
        shards: dict
            A dict of shard definitions where key is the shard's name and value is the host/database string.
        user: str
        password: str
        timeout: int
            Statement timeout in milliseconds.
        shard: str
            Optional shard name. If provided, the check will be run on only one shard matching given name.
        '''

        if not shards:
            raise ValueError('SqlWrapper: No shards defined')
        if shard and not shards.get(shard):
            raise ValueError('SqlWrapper: Shard {} not found in shards definition'.format(shard))

        self._cursors = []
        self._stmt = None

        mdb = _import_db_driver()

        for shard_def in ([shards[shard]] if shard else shards.values()):
            m = CONNECTION_RE.match(shard_def)
            if not m:
                raise ValueError('Invalid shard connection: {}'.format(shard_def))
            try:
                conn = mdb.connect(host=m.group('host'), user=user, passwd=password, db=m.group('dbname'),
                                   port=(int(m.group('port')) if int(m.group('port')) > 0 else DEFAULT_PORT),
                                   connect_timeout=timeout)
            except Exception, e:
                raise DbError(str(e), operation='Connect to {}'.format(shard_def)), None, sys.exc_info()[2]

            # TODO: Find a way to enforce readonly=True as it is done in postgres Wrapper
            # TODO: Do we need to set charset="utf8" and use_unicode=True in connection?

            conn.autocommit(True)
            self._cursors.append(conn.cursor(mdb.cursors.DictCursor))

    def execute(self, stmt):
        self._stmt = stmt
        return self

    def result(self, agg=sum):
        '''return single row result, will result primitive value if only one column is selected'''

        result = {}
        try:
            for cur in self._cursors:
                try:
                    cur.execute(self._stmt)
                    row = cur.fetchone()
                    if row:
                        for k, v in row.items():
                            result[k] = result.get(k, [])
                            result[k].append(v)
                finally:
                    cur.close()
        except Exception, e:
            raise DbError(str(e), operation=self._stmt), None, sys.exc_info()[2]

        for k, v in result.items():
            try:
                result[k] = agg(v)
            except:
                # just use list if aggregation function fails
                # (e.g. if we try to sum strings)
                result[k] = v
        if len(result) == 1:
            return result.values()[0]
        else:
            return result

    def results(self):
        '''return many rows'''

        results = []
        try:
            for cur in self._cursors:
                try:
                    cur.execute(self._stmt)
                    rows = cur.fetchmany(MAX_RESULTS)
                    for row in rows:
                        results.append(dict(row))
                finally:
                    cur.close()
        except Exception, e:
            raise DbError(str(e), operation=self._stmt), None, sys.exc_info()[2]
        return results


if __name__ == '__main__':
    default_dbname = 'mysql'
    default_sql_stmt = 'SELECT VERSION()'  # SELECT Host,User FROM user

    if len(sys.argv) in (6, 7):
        if len(sys.argv) == 7:
            shards = {'test': sys.argv[1] + ':' + sys.argv[2] + '/' + sys.argv[3]}
            sql_stmt = sys.argv[6]
        else:
            print 'executing default statement:', default_sql_stmt
            shards = {'test': sys.argv[1] + ':' + sys.argv[2] + '/' + default_dbname}
            sql_stmt = default_sql_stmt

        check = MySqlWrapper(shards, user=sys.argv[4], password=sys.argv[5])
        # print '>>> many results:\n', check.execute(sql_stmt).results()
        print '>>> one result:\n', check.execute(sql_stmt).result()
    else:
        print '{} <host> <port> <dbname> <user> <password> [sql_stmt]'.format(sys.argv[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class CheckError(Exception):

    pass


class SecurityError(Exception):

    pass


class NagiosError(CheckError):

    def __init__(self, output):
        self.output = output
        super(NagiosError, self).__init__()

    def __str__(self):
        return 'NagiosError. Command output: {}'.format(self.output)


class InsufficientPermissionsError(CheckError):

    def __init__(self, user, entity):
        self.user = user
        self.entity = entity

    def __str__(self):
        return 'Insufficient permisions for user {} to access {}'.format(self.user, self.entity)


class SnmpError(CheckError):

    def __init__(self, message):
        self.message = message
        super(SnmpError, self).__init__()

    def __str__(self):
        return 'SNMP Error. Message: {}'.format(self.message)


class JmxQueryError(CheckError):

    def __init__(self, message):
        self.message = message
        super(JmxQueryError, self).__init__()

    def __str__(self):
        return 'JMX Query failed: {}'.format(self.message)


class HttpError(CheckError):

    def __init__(self, message, url=None):
        self.message = message
        self.url = url
        super(HttpError, self).__init__()

    def __str__(self):
        return 'HTTP request failed for {}: {}'.format(self.url, self.message)


class DbError(CheckError):

    def __init__(self, message, operation=None):
        self.message = message
        self.operation = operation
        super(DbError, self).__init__()

    def __str__(self):
        return 'DB operation {} failed: {}'.format(self.operation, self.message)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
We provide a plugin system to load external functionality into Zmon.
This is needed as a way to decouple sensitive details -possibly proprietary- from the Zmon core.

The plugin system is separated in *adapters* and *implementations*.

Adapters are base classes that specify the behaviour of plugin types, they live in the subpackage
``zmon_worker.adapters``. All apaters inherit from ``zmon_worker.adapters.IBasePlugin``.
One adapter may have many implementations,

A plugin implementation needs 2 files:
1. a python source file containing a class that extends an adaptor.
2. a plugin info file. Text file with metadata to uniquely identify the plugin.


To see how this works let's look at the *resolution of the entities* to monitor:

What Zmon scheduler does is basically to periodically run some checks against some entities. An entity may be a host,
a database, a network port, or really anything you may want to run a check against. This means that the definition of
**entity** must be kept open and flexible, that's why the user must implement an entity plugin to define its entities.

How do you write an entity plugin to load your entities (hosts, databases, network ports, etc.)?

1. Write a python file containing a class that extends ``zmon_worker.adapters.ientity_plugin.IEntityPlugin``,
this class should contain your own logic for resolving your entities.

2. Write a text file with the same name of your python source but extension ``.scheduler_plugin``.

3. Place both files in a *plugin folder* of your choice and add that folder's absolute path to the
environment variable ``ZMON_PLUGINS``.


This description is quite dry, but we provide an example that will make things clearer:

The scheduler comes with one functional plugin, called ``entity_city``, this plugin provides *city entities*, these
are used for internal Zmon checks. This plugin is a simple template to start external plugin implementation.
The 2 files that implement the city plugin are ``zmon_scheduler.builtins.plugins.city.scheduler_plugin`` and
``zmon_scheduler.builtins.plugins.city.py``.

"""

from zmon_worker_monitor.adapters.ibase_plugin import IBasePlugin


__all__ = [
    'IBasePlugin',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dogpile.cache import make_region
import threading
from functools import wraps
import sys
import os
import time
import random

DEFAULT_CACHE_EXPIRATION_TIME = 3600


def get_cache_file_path():
    cache_filename = 'worker_cache.dbm'
    app_folder = (os.environ['APP_HOME'] if 'APP_HOME' in os.environ else os.getcwd())
    cache_dir = (os.path.join(app_folder, 'temp') if os.path.isdir(os.path.join(app_folder, 'temp')) else app_folder)
    return os.path.join(cache_dir, cache_filename)


def async_creation_runner(cache, somekey, creator, mutex):
    ''' Used by dogpile.core:Lock when appropriate '''

    def runner():
        try:
            value = creator()
            cache.set(somekey, value)
        finally:
            mutex.release()

    thread = threading.Thread(target=runner)
    thread.start()


# Asynchronous cache decorator persisted to memory.
# After the first successful invocation cache updates happen in a background thread.
async_memory_cache = make_region(async_creation_runner=async_creation_runner).configure('dogpile.cache.memory',
        expiration_time=DEFAULT_CACHE_EXPIRATION_TIME)


def with_retries(max_retries=5, delay=5):
    ''' A helper decorator to handle retries on function call failures '''

    def decorator(f):

        random_delta = 0.25  # percent of random deviation we will introduce in waiting time

        @wraps(f)
        def wrapper(*args, **kwargs):
            retry, ts = 0, delay
            while True:
                try:
                    return f(*args, **kwargs)
                except Exception, e:
                    if retry >= max_retries:
                        traceback = sys.exc_info()[2]
                        raise Exception('Max retries exceeded. Internal error: {}'.format(e)), None, traceback
                    time.sleep(ts + random.uniform(-ts, ts) * random_delta)
                    retry += 1
                    ts *= 2  # duplicate waiting time
        return wrapper

    return decorator





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import psycopg2
import re
import sys

from zmon_worker_monitor.zmon_worker.errors import CheckError, InsufficientPermissionsError, DbError
from psycopg2.extras import NamedTupleCursor

DEFAULT_PORT = 5432

CONNECTION_RE = \
    re.compile(r'''
^(?P<host>[^:/]+)       # host - either IP o hostname
(:(?P<port>\d+))?       # port - integer, optional
/(?P<dbname>\w+)        # database name
$
'''
               , re.X)

ABSOLUTE_MAX_RESULTS = 1000000

REQUIRED_GROUP = 'zalandos'
PERMISSIONS_STMT = \
    '''
SELECT r.rolcanlogin AS can_login,
       ARRAY(SELECT b.rolname
               FROM pg_catalog.pg_auth_members m
               JOIN pg_catalog.pg_roles b ON (m.roleid = b.oid)
              WHERE m.member = r.oid) AS member_of
  FROM pg_catalog.pg_roles r
 WHERE r.rolname = %s;
'''

NON_SAFE_CHARS = re.compile(r'[^a-zA-Z_0-9-]')


def make_safe(s):
    '''
    >>> make_safe('Bad bad \\' 123')
    'Badbad123'
    '''

    if not s:
        return ''
    return NON_SAFE_CHARS.sub('', s)


class SqlWrapper(object):

    '''Shard-aware SQL adapter
    sql().execute('SELECT 1').result()
    '''

    def __init__(
        self,
        shards,
        user='nagios',
        password='',
        timeout=60000,
        shard=None,
        created_by=None,
        check_id=None,
    ):
        '''
        Parameters
        ----------
        shards: dict
            A dict of shard definitions where key is the shard's name and value is the host/database string.
        user: str
        password: str
        timeout: int
            Statement timeout in milliseconds.
        shard: str
            Optional shard name. If provided, the check will be run on only one shard matching given name.
        created_by: str
            Optional user name. If provided, the check will first make sure that the user has permissions to access the
            requested database. It's optional because it's currently supported only in trial run.
        check_id: int
            The check definition ID in order to set PostgreSQL application name (easier tracking on server side).
        '''

        if not shards:
            raise CheckError('SqlWrapper: No shards defined')
        if shard and not shards.get(shard):
            raise CheckError('SqlWrapper: Shard {} not found in shards definition'.format(shard))

        self._cursors = []
        self._stmt = None
        permissions = {}

        for shard_def in ([shards[shard]] if shard else shards.values()):
            m = CONNECTION_RE.match(shard_def)
            if not m:
                raise CheckError('Invalid shard connection: {}'.format(shard_def))
            connection_str = \
                "host='{host}' port='{port}' dbname='{dbname}' user='{user}' password='{password}' connect_timeout=5 options='-c statement_timeout={timeout}' application_name='ZMON Check {check_id} (created by {created_by})' ".format(
                host=m.group('host'),
                port=int(m.group('port') or DEFAULT_PORT),
                dbname=m.group('dbname'),
                user=user,
                password=password,
                timeout=timeout,
                check_id=check_id,
                created_by=make_safe(created_by),
            )
            try:
                conn = psycopg2.connect(connection_str)
                conn.set_session(readonly=True, autocommit=True)
                cursor = conn.cursor(cursor_factory=NamedTupleCursor)
                self._cursors.append(cursor)
            except Exception, e:
                raise DbError(str(e), operation='Connect to {}'.format(shard_def)), None, sys.exc_info()[2]
            try:
                if created_by:
                    cursor.execute(PERMISSIONS_STMT, [created_by])
                    row = cursor.fetchone()
                    permissions[shard_def] = (row.can_login and REQUIRED_GROUP in row.member_of if row else False)
            except Exception, e:
                raise DbError(str(e), operation='Permission query'), None, sys.exc_info()[2]

        for resource, permitted in permissions.iteritems():
            if not permitted:
                raise InsufficientPermissionsError(created_by, resource)

    def execute(self, stmt):
        self._stmt = stmt
        return self

    def result(self, agg=sum):
        '''return single row result, will result primitive value if only one column is selected'''

        result = {}
        try:
            for cur in self._cursors:
                try:
                    cur.execute(self._stmt)
                    row = cur.fetchone()
                    if row:
                        for k, v in row._asdict().items():
                            result[k] = result.get(k, [])
                            result[k].append(v)
                finally:
                    cur.close()
        except Exception, e:
            raise DbError(str(e), operation=self._stmt), None, sys.exc_info()[2]

        for k, v in result.items():
            try:
                result[k] = agg(v)
            except:
                # just use list if aggregation function fails
                # (e.g. if we try to sum strings)
                result[k] = v
        if len(result) == 1:
            return result.values()[0]
        else:
            return result

    def results(self, max_results=100, raise_if_limit_exceeded=True):
        '''return many rows'''

        results = []
        max_results = min(max_results, ABSOLUTE_MAX_RESULTS)
        try:
            for cur in self._cursors:
                try:
                    cur.execute(self._stmt)
                    if raise_if_limit_exceeded:
                        rows = cur.fetchmany(max_results + 1)
                        if len(rows) > max_results:
                            raise DbError('Too many results, result set was limited to {}. Try setting max_results to a higher value.'.format(max_results),
                                          operation=self._stmt)
                    else:
                        rows = cur.fetchmany(max_results)
                    for row in rows:
                        results.append(row._asdict())
                finally:
                    cur.close()
        except Exception, e:
            raise DbError(str(e), operation=self._stmt), None, sys.exc_info()[2]
        return results


if __name__ == '__main__':
    if len(sys.argv) == 4:
        check = SqlWrapper([sys.argv[1] + '/' + sys.argv[2]])
        print check.execute(sys.argv[3]).result()
    elif len(sys.argv) > 1:
        print 'sql.py <host> <dbname> <stmt>'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime

from zmon_worker_monitor.zmon_worker.common.time_ import parse_timedelta, parse_datetime

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial

class TimeFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(TimeFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(TimeWrapper)

class TimeWrapper(object):

    def __init__(self, spec='now', utc=False):
        now = (datetime.utcnow() if utc else datetime.now())
        delta = parse_timedelta(spec)
        if delta:
            self.time = now + delta
        elif spec == 'now':
            self.time = now
        else:
            self.time = parse_datetime(spec)

    def __sub__(self, other):
        '''
        >>> TimeWrapper('2014-01-01 01:01:25') - TimeWrapper('2014-01-01 01:01:01')
        24.0
        '''

        return (self.time - other.time).total_seconds()

    def isoformat(self, sep=' '):
        return self.time.isoformat(sep)

    def format(self, fmt):
        '''
        >>> TimeWrapper('2014-01-01 01:01').format('%Y-%m-%d')
        '2014-01-01'

        >>> TimeWrapper('-1m').format('%Y-%m-%d')[:3]
        '201'
        '''

        return self.time.strftime(fmt)



import eventlog
import json
import requests
import datetime
from eventlog import Event

_target_host = 'localhost'
_target_port = 8081
_enable_http = True

def set_target_host(host='localhost', port='8081'):
    global _target_host, _target_port
    _target_host = host
    _target_port = port

def enable_http(enable=True):
    global _enable_http
    _enable_http = enable

def register_all(events, path=None):
    eventlog.register_all(events, path)

def log(e_id, **kwargs):
    # for now forward everything
    eventlog.log(e_id, **kwargs)

    if not _enable_http:
        return

    now = datetime.datetime.now()

    headers = {'content-type': 'application/json'}

    event = {'typeId': e_id, 'attributes': kwargs, 'time': now.strftime("%Y-%m-%dT%H:%M:%S.")+now.strftime("%f")[:3]}

    try:
        res = requests.put('http://{}:{}/'.format(_target_host, _target_port), data=json.dumps([event]), headers=headers)
    except Exception as e:
        pass




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from ConfigParser import ConfigParser
import settings

from zmon_worker.tasks.notacelery_task import NotaZmonTask
from zmon_worker.notifications.mail import Mail
from zmon_worker.notifications.sms import Sms


logger = logging.getLogger(__name__)


def load_config_from_file():
    # load the global section of the configuration as a dict (eval to get values as python objects)
    c = ConfigParser()
    c.readfp(open(settings.zmon_worker_config_file))
    config = dict((key, eval(val)) for key, val in c.items('global'))

    logger.info('loaded worker config is: %s', config)

    return config


def configure_tasks(config):
    #Pass configuration to zmon classes
    NotaZmonTask.configure(config)

    Mail.update_config(config)
    Sms.update_config(config)


zmontask = NotaZmonTask()


def check_and_notify(req, alerts, task_context=None, **kwargs):
    logger.debug('check_and_notify received req=%s, alerts=%s, task_context=%s, ', req, alerts, task_context)

    zmontask.check_and_notify(req, alerts, task_context=task_context)


def trial_run(req, alerts, task_context=None, **kwargs):
    logger.info('trial_run received <== check_id=%s', req['check_id'])
    logger.debug('trial_run received req=%s, alerts=%s, task_context=%s, ', req, alerts, task_context)

    zmontask.trial_run(req, alerts, task_context=task_context)


def cleanup(*args, **kwargs):
    logger.info('cleanup task received with args=%s, kwargs=%s', args, kwargs)

    zmontask.cleanup(*args, **kwargs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy
import datetime
from zmon_worker_monitor.zmon_worker.common.time_ import parse_timedelta


# from tasks.check import flatten

# originally, I wanted to load this function defintion from the tasks module, but I did not
# succeed in doing so. My python knowledge is limited, so maybe you can tell me how I can achieve this?

def flatten(structure, key='', path='', flattened=None):
    path = str(path)
    key = str(key)

    if flattened is None:
        flattened = {}
    if type(structure) not in (dict, list):
        flattened[((path + '.' if path else '')) + key] = structure
    elif isinstance(structure, list):
        pass
    else:
        #        for i, item in enumerate(structure):
        #            flatten(item, '%d' % i, '.'.join(filter(None, [path, key])), flattened)
        for new_key, value in structure.items():
            flatten(value, new_key, '.'.join(filter(None, [path, key])), flattened)
    return flattened


class DistanceWrapper(object):

    def __init__(self, history_wrapper, weeks=4, snap_to_bin=True, bin_size='1h', dict_extractor_path=''):

        self.history_wrapper = history_wrapper
        self.weeks = weeks
        self.snap_to_bin = snap_to_bin
        self.bin_size = parse_timedelta(bin_size)
        self.dict_extractor_path = dict_extractor_path

    def calculate_bin_time_range(self):
        '''
        Calculates the time ranges we need to look for using the settings configured for this class.
        Returns a list of dicts with the time ranges.
        '''

        now = datetime.datetime.now().replace(microsecond=0)
        if self.snap_to_bin:
            day_begin = now.replace(hour=0, minute=0, second=0, microsecond=0)
            # the number of bins (of size bin_size) that passed since the beginning of the day
            bins_this_day_until_now = int((now - day_begin).total_seconds() / self.bin_size.total_seconds())
            bin_begin = day_begin + bins_this_day_until_now * self.bin_size
            bin_end = day_begin + (bins_this_day_until_now + 1) * self.bin_size
        else:
            bin_begin = now - self.bin_size
            bin_end = now

        timestamps = []
        for week in range(1, self.weeks + 1):
            time_from = abs((bin_begin - week * datetime.timedelta(days=7) - now).total_seconds())
            time_to = abs((bin_end - week * datetime.timedelta(days=7) - now).total_seconds())
            timestamps.append({'time_from': time_from, 'time_to': time_to})
        return timestamps

    def extract_value(self, value):
        '''
        Extracts the value that will be used for comparisons. For dicts, we need an extractor. The extractor may not
        be empty if the value is a dict. The extractor is of the form 'a.b.c' for a dict with the structure
        {'a':{'b':{'c':5}}} to extract the value 5.
        '''

        if isinstance(value, dict):
            if self.dict_extractor_path == '':
                raise Exception('Extractor may not be empty for dicts as value. You need to tell which part of the dict I should use.'
                                )
            # throws when the key is unavailable; this is what we want
            value = flatten(value)[self.dict_extractor_path]
        return value

    def bin_mean(self):
        '''
        Calculates the mean of the history values. Applies the extractor and returns a scalar value.
        '''

        time_ranges = self.calculate_bin_time_range()
        averages = []
        for time_range in time_ranges:
            averages.extend(self.history_wrapper.get_avg(self.dict_extractor_path, time_range['time_from'],
                            time_range['time_to']))
        if not averages:
            raise Exception('No history data available in bin_mean call.')
        return numpy.average(averages)

    def bin_standard_deviation(self):
        '''
        Calculates the standard deviation of the history values. Applies the extractor and returns a scalar value.
        '''

        time_ranges = self.calculate_bin_time_range()
        deviations = []
        for time_range in time_ranges:
            deviations.extend(self.history_wrapper.get_std_dev(self.dict_extractor_path, time_range['time_from'],
                              time_range['time_to']))
        if not deviations:
            raise Exception('No history data available in bin_standard_deviation call.')

        # you can't simply average standard deviations.
        # see https://en.wikipedia.org/wiki/Variance#Basic_properties for details, keep in mind that
        # the different times are uncorrelated. We assume that the sample sizes for the different weeks
        # are equal (since we do not get exact sample sizes for a specific key from the kairosdb)
        return numpy.sqrt(numpy.sum(map(lambda x: x * x, deviations)))

    def absolute(self, value):
        '''
        Calculates the absolute distance of the actual value from the history value bins
        as selected through the constructor parameters weeks, bin_size and snap_to_bin.
        Applies the extractor and returns a scalar value.
        '''

        return self.extract_value(value) - self.bin_mean()

    def sigma(self, value):
        '''
        Calculates the relative distance of the actual value from the history value bins,
        normalized by the standard deviation. A sigma distance of 1.0 means that the actual value
        is as far away from the mean as the standard deviation is. The sigma distance can be negative, in this
        case you are below the mean with your value. If you need absolute values, you can use
        abs(sigma(value)). Applies the extractor and returns a scalar value.
        '''

        abs_value = self.absolute(value)
        std_dev = self.bin_standard_deviation()

        if std_dev == 0:
            return numpy.Infinity * numpy.sign(abs_value) if abs_value != 0 else numpy.float64(0)

        return abs_value / std_dev


if __name__ == '__main__':
    import logging

    logging.basicConfig(level=logging.DEBUG)

    logging.info('flattened dict: %s', flatten({'a': {'b': {'c': 5}}}))


    class HistoryWrapper(object):

        def __init__(self, check_id, entities=[]):
            self.check_id = check_id
            self.entities = entities

        @staticmethod
        def get_avg(key, time_from, time_to):
            return [7]

        @staticmethod
        def get_std_dev(key, time_from, time_to):
            return [2]


    now = datetime.datetime.now()
    before = now.replace(minute=0, second=0, microsecond=0)

    distance = DistanceWrapper(history_wrapper=HistoryWrapper(check_id=588), snap_to_bin=False, weeks=3, bin_size='5m')
    logging.info('Mean of history values: %f', distance.bin_mean())
    logging.info('Absolute distance: %f', distance.absolute(15))
    logging.info('Sigma distance: %f', distance.sigma(15))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# wrapper for kairosdb to access history data about checks

import logging
import requests

from zmon_worker_monitor.builtins.plugins.distance_to_history import DistanceWrapper

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


logger = logging.getLogger(__name__)


class HistoryFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(HistoryFactory, self).__init__()
        # fields from configuration
        self.kairosdb_host = None
        self.kairosdb_port = None
        self.kairosdb_history_enabled = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self.kairosdb_host = conf.get('kairosdb_host')
        self.kairosdb_port = conf.get('kairosdb_port')
        self.kairosdb_history_enabled = True if conf.get('kairosdb_history_enabled') in ('true', 'True', '1') else False

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(HistoryWrapper,
                          kairosdb_host=self.kairosdb_host,
                          kairosdb_port=self.kairosdb_port,
                          history_enabled=self.kairosdb_history_enabled,
                          check_id=factory_ctx['check_id'],
                          entities=factory_ctx['entity_id_for_kairos'])


def get_request_json(check_id, entities, time_from, time_to, aggregator='avg', sampling_size_in_seconds=300):
    j = \
        """
{
  "metrics": [
    {
      "tags": {
        "entity": [
          %s
        ]
      },
      "name": "zmon.check.%s",
      "group_by": [
        {
          "name": "tag",
          "tags": [
            "key"
          ]
        }
      ],
      "aggregators": [
        {
          "name": "%s",
          "align_sampling": true,
          "sampling": {
            "value": "%s",
            "unit": "seconds"
          }
        }
      ]
    }
  ],
  "cache_time": 0,
  "start_relative": {
    "value": "%s",
    "unit": "seconds"
  },
  "end_relative": {
    "value": "%s",
    "unit": "seconds"
  }
}
"""

    r = j % (
        ','.join(map(lambda x: '"' + x + '"', entities)),
        check_id,
        aggregator,
        sampling_size_in_seconds,
        time_from,
        time_to,
    )
    return r


ONE_WEEK = 7 * 24 * 60 * 60
ONE_WEEK_AND_5MIN = ONE_WEEK + 5 * 60


class HistoryWrapper(object):

    def __init__(self, kairosdb_host, kairosdb_port, history_enabled, check_id, entities):

        self.__kairosdb_host = kairosdb_host if kairosdb_host is not None else 'cassandra01'
        self.__kairosdb_port = kairosdb_port if kairosdb_port is not None else '37629'
        self.__enabled = bool(history_enabled)

        self.url = 'http://%s:%s/api/v1/datapoints/query' % (self.__kairosdb_host, self.__kairosdb_port)
        self.check_id = check_id

        if type(entities) == list:
            self.entities = entities
        else:
            self.entities = [entities]

    def result(self, time_from=ONE_WEEK_AND_5MIN, time_to=ONE_WEEK):
        if not self.__enabled:
            raise Exception("History() function disabled for now")

        #self.logger.info("history query %s %s %s", self.check_id, time_from, time_to)
        return requests.post(self.url, get_request_json(self.check_id, self.entities, int(time_from),
                             int(time_to))).json()

    def get_one(self, time_from=ONE_WEEK_AND_5MIN, time_to=ONE_WEEK):
        if not self.__enabled:
            raise Exception("History() function disabled for now")

        #self.logger.info("history get one %s %s %s", self.check_id, time_from, time_to)
        return requests.post(self.url, get_request_json(self.check_id, self.entities, int(time_from),
                             int(time_to))).json()['queries'][0]['results'][0]['values']

    def get_aggregated(self, key, aggregator, time_from=ONE_WEEK_AND_5MIN, time_to=ONE_WEEK):
        if not self.__enabled:
            raise Exception("History() function disabled for now")

        # read the list of results
        query_result = requests.post(self.url, get_request_json(self.check_id, self.entities, int(time_from),
                                     int(time_to), aggregator, int(time_from - time_to))).json()['queries'][0]['results'
                ]

        # filter for the key we are interested in
        filtered_for_key = filter(lambda x: x['tags'].get('key', [''])[0] == key, query_result)

        if not filtered_for_key:
            return_value = []
        else:
            return_value = [filtered_for_key[0]['values'][0][1]]

        # since we have a sample size of 'all in the time range', return only the value, not the timestamp.
        return return_value

    def get_avg(self, key, time_from=ONE_WEEK_AND_5MIN, time_to=ONE_WEEK):
        if not self.__enabled:
            raise Exception("History() function disabled for now")

        #self.logger.info("history get avg %s %s %s", self.check_id, time_from, time_to)
        return self.get_aggregated(key, 'avg', time_from, time_to)

    def get_std_dev(self, key, time_from=ONE_WEEK_AND_5MIN, time_to=ONE_WEEK):
        if not self.__enabled:
            raise Exception("History() function disabled for now")

        #self.logger.info("history get std %s %s %s", self.check_id, time_from, time_to)
        return self.get_aggregated(key, 'dev', time_from, time_to)

    def distance(self, weeks=4, snap_to_bin=True, bin_size='1h', dict_extractor_path=''):
        if not self.__enabled:
            raise Exception("History() function disabled for now")

        #self.logger.info("history distance %s %s ", self.check_id, weeks, bin_size)
        return DistanceWrapper(history_wrapper=self, weeks=weeks, bin_size=bin_size, snap_to_bin=snap_to_bin,
                               dict_extractor_path=dict_extractor_path)


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.DEBUG)
    zhistory = HistoryWrapper(None, None, None, 17, ['GLOBAL'])
    r = zhistory.result()
    logging.info(r)
    r = zhistory.get_one()
    logging.info(r)
    r = zhistory.get_aggregated('', 'avg')
    logging.info(r)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from zmon_worker_monitor.zmon_worker.errors import CheckError
from functools import partial
from suds.client import Client
from zmon_worker_monitor.zmon_worker.common.time_ import parse_timedelta
from timeperiod import in_period, InvalidFormat
from zmon_worker_monitor.zmon_worker.common.utils import async_memory_cache

import sys
import json
import redis
import time

logger = logging.getLogger(__name__)

CHECK_REFRESH_TIME = 240
ALERT_REFRESH_TIME = 120


class ZmonWrapper(object):

    ZMON_ALERTS_ENTITIES_PATTERN = 'zmon:alerts:*:entities'

    def __init__(self, wsdl, host, port):
        try:
            self.__ws_client = Client(url=wsdl)
            self.__ws_client.set_options(cache=None)
        except Exception:
            raise CheckError('ZmonWrapper Error: failed to connect to zmon-controller')

        self.__redis = redis.StrictRedis(host, port)
        self.__checks = {}
        self.__alerts = []

        self.logger = logger

        self.__checks = self.__load_check_definitions()
        self.__alerts = self.__load_alert_definitions()

    @async_memory_cache.cache_on_arguments(namespace='zmon-worker', expiration_time=ALERT_REFRESH_TIME)
    def __load_alert_definitions(self):
        try:
            response = self.__ws_client.service.getAllActiveAlertDefinitions()
        except Exception:
            self.logger.exception('ZmonWrapper Error: failed to load alert definitions')
            raise CheckError('ZmonWrapper Error: failed to load alert definitions'), None, sys.exc_info()[2]
        else:
            return [{
                'id': a.id,
                'team': a.team,
                'responsible_team': a.responsibleTeam,
                'check_id': a.checkDefinitionId,
                'period': (a.period or '' if hasattr(a, 'period') else ''),
            } for a in response[1]]

    @async_memory_cache.cache_on_arguments(namespace='zmon-worker', expiration_time=CHECK_REFRESH_TIME)
    def __load_check_definitions(self):
        try:
            response = self.__ws_client.service.getAllActiveCheckDefinitions()
        except Exception:
            self.logger.exception('ZmonWrapper Error: failed to load check definitions')
            raise CheckError('ZmonWrapper Error: failed to load check definitions'), None, sys.exc_info()[2]
        else:
            return dict((c.id, {'interval': c.interval}) for c in response[1])

    @staticmethod
    def _is_entity_alert_stale(last_run, period):
        '''
        Checks whether check's last run is within given period.

        >>> ZmonWrapper._is_entity_alert_stale(None, 60)
        False
        >>> ZmonWrapper._is_entity_alert_stale(time.time(), 10)
        False
        >>> ZmonWrapper._is_entity_alert_stale(time.time() - 20, 10)
        True
        '''

        return (False if last_run is None else time.time() - last_run > period)

    def __is_alert_stale(self, alert, evaluated_alerts, check_results, multiplier, offset):
        a_id = alert['id']  # alert id
        c_id = alert['check_id']  # check id
        r_id = partial('{}:{}'.format, c_id)  # helper function used in iterator to generate result id

        try:
            is_in_period = in_period(alert.get('period', ''))
        except InvalidFormat:
            self.logger.warn('Alert with id %s has malformed time period.', a_id)
            is_in_period = True

        if is_in_period:
            return a_id not in evaluated_alerts or any(self._is_entity_alert_stale(check_results.get(r_id(entity)),
                                                       multiplier * self.__checks[c_id]['interval'] + offset)
                                                       for entity in evaluated_alerts[a_id])
        else:
            return False

    def stale_active_alerts(self, multiplier=2, offset='5m'):
        '''
        Returns a list of alerts that weren't executed in a given period of time. The period is calculated using
        multiplier and offset: check's interval * multiplier + offset.
        Parameters
        ----------
        multiplier: int
            Multiplier for check's interval.
        offset: str
            Time offset, for details see parse_timedelta function in zmon-worker/src/function/time_.py.
        Returns
        -------
        list
            A list of stale active alerts.
        '''

        alert_entities = self.__redis.keys(self.ZMON_ALERTS_ENTITIES_PATTERN)

        # Load evaluated alerts and their entities from redis.
        p = self.__redis.pipeline()
        for key in alert_entities:
            p.hkeys(key)
        entities = p.execute()
        evaluated_alerts = dict((int(key.split(':')[2]), entities[i]) for (i, key) in enumerate(alert_entities))

        # Load check results for previously loaded alerts and entities.
        check_ids = []
        for alert in self.__alerts:
            if alert['id'] in evaluated_alerts:
                for entity in evaluated_alerts[alert['id']]:
                    p.lindex('zmon:checks:{}:{}'.format(alert['check_id'], entity), 0)
                    check_ids.append('{}:{}'.format(alert['check_id'], entity))
        results = p.execute()
        check_results = dict((check_id, json.loads(results[i])['ts']) for (i, check_id) in enumerate(check_ids)
                             if results[i])

        return [{'id': alert['id'], 'team': alert['team'], 'responsible_team': alert['responsible_team']} for alert in
                self.__alerts if self.__is_alert_stale(alert, evaluated_alerts, check_results, multiplier,
                parse_timedelta(offset).total_seconds())]

    def check_entities_total(self):
        '''
        Returns total number of checked entities.
        '''

        alert_entities = self.__redis.keys(self.ZMON_ALERTS_ENTITIES_PATTERN)
        p = self.__redis.pipeline()
        for key in alert_entities:
            p.hkeys(key)
        entities = p.execute()

        return sum(len(e) for e in entities)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from scmmanager import StashApi
except:
    StashApi = None
import os
import yaml
import logging
from yaml.scanner import ScannerError


class StashAccessor(object):

    def __init__(self, logger=None):
        self._logger = (logger if logger else _get_null_logger())
        if StashApi:
            self.stash = StashApi()
        else:
            self.stash = None

    def set_logger(self, logger):
        self._logger = logger

    def get_stash_check_definitions(self, *repo_urls):
        '''
        Downloads the check definitions found under the given stash repositories.
        Returns a list of dict corresponding to the check definitions
        '''

        check_definitions = []
        if self.stash:
            for repo_url in repo_urls:
                self._logger.info("Scanning Stash url: %s", repo_url)
                try:
                    repo, check_path = self.stash.get_repo_from_scm_url(repo_url)
                except Exception:
                    self._logger.exception('Bad configured stash repo: %s. Exception follows: ', repo_url)
                else:
                    try:
                        files = [f for f in self.stash.get_files(repo, check_path) if f.endswith('.yaml')]
                        if not files:
                            self._logger.warn('No check definitions found in secure repo: %s', repo_url)

                        for check_file in files:
                            self._logger.info("Reading file: %s", check_file)

                            file_path = os.path.join(check_path, check_file)
                            fd = self.stash.get_content(repo, file_path)

                            try:
                                check_definitions.append(yaml.safe_load(fd))
                            except ScannerError as e:
                                self._logger.exception("Could not parse file %s/%s", check_path, check_file, e )

                    except Exception:
                        self._logger.exception('Unexpected error when fetching info from stash: ')

            if repo_urls and not check_definitions:  # StashApi returns empty results on failure
                self._logger.error('Stash error: Check definition download finished with empty results')
                raise Exception('Check definition download finished with empty results')

        return check_definitions

    def get_stash_commands(self, *repo_urls):
        '''
        Returns a list of str corresponding to commands found in check_definitions under the given stash repositories
        '''

        return set(cf['command'] for cf in self.get_stash_check_definitions(*repo_urls) if 'command' in cf)


class NullHandler(logging.Handler):

    def emit(self, record):
        pass


_null_logger = None


def _get_null_logger():
    global _null_logger
    if not _null_logger:
        handler = NullHandler()
        _null_logger = logging.getLogger('NullLogger')
        _null_logger.addHandler(handler)
    return _null_logger





#!/usr/bin/env python
# -*- coding: utf-8 -*-

from emu_kombu import parse_redis_conn
import redis
import os
import logging
import time
from threading import current_thread
from threading import local as thread_local
import collections
import math
from traceback import format_exception


logger = logging.getLogger(__name__)


WAIT_RECONNECT_MIN = 0.1
WAIT_RECONNECT_MAX = 20


class _ThreadLocal(thread_local):
    can_init = False
    instance = None


class RedisConnHandler(object):

    """
    This is a connection manager for redis implemented as a context handler. When used inside a with statement
    it intercepts RedisConnection exceptions as well as its own IdleLoopException in order to keep score of
    failures and idle cycles. Based in this counters it reacts to connection errors by introducing small exponential
    time delays and making several attempts to regain the connection, if t_wait_per_server seconds pass without
    success it switches to the next redis server from the list it given when configured.
    It also switches to the next server after t_wait_no_tasks seconds without getting any task.
    You also get thread safety (connections are not shared among threads), and an easy way to get the connection
    without passing the reference around.
    """

    # Constants
    __CONS_SUPPRESS_EXCEPTION = True

    __CONS_PROPAGATE_EXCEPTION = False

    STATUS_ERROR = 'STATUS_ERROR'
    STATUS_IDLE = 'STATUS_IDLE'
    STATUS_OK = 'STATUS_OK'

    # class variables
    servers = []

    t_wait0 = WAIT_RECONNECT_MIN

    reties_per_server = 5

    t_wait_per_server = 30  # if 30 seconds pass and we have connection errors we switch server

    t_wait_no_tasks = 5 * 60  # if 5 minutes pass without getting any message we switch server

    #t_reconnect_master = 10 * 60  # in 10 minutes it will attempt to connect to server 0 again

    _pid = None

    _max_wait_step = 15  # a top value for our exponential increase in waiting time

    _thread_local = _ThreadLocal()

    # Counters and dates markers for connection errors
    _active_index = 0
    _retries_count = -1
    _idle_count = -1
    message_count = 0
    _last_failure_tstamp = 0
    _last_success_tstamp = time.time()
    _last_message_tstamp = time.time()


    class IdleLoopException(Exception):
        pass

    @classmethod
    def configure(cls, **config):

        cls._pid = os.getpid()

        cls.t_wait0 = float(config.get('t_wait0', cls.t_wait0))

        cls.t_wait_no_tasks = float(config.get('t_wait_no_tasks', cls.t_wait_no_tasks))

        cls.t_wait_per_server = float(config.get('t_wait_per_server', cls.t_wait_per_server))

        # estimate the reties_per_server from the wait_time_per_server
        cls.reties_per_server = cls.calculate_reties_per_server(cls.t_wait_per_server, cls.t_wait0)

        #cls.t_reconnect_master = int(config.get('t_reconnect_master', cls.t_reconnect_master))

        servers = config.get('redis_servers')
        if servers:
            if isinstance(servers, basestring):
                servers = [s.strip() for s in servers.split(',')]
            elif not isinstance(servers, collections.Iterable):
                raise Exception("wrong servers parameter")
            cls.servers = list(servers)
            # parse all server urls to detect config errors beforehand
            [parse_redis_conn(s) for s in cls.servers]

        logger.warn('Pid=%s ==> RedisConnHandler  configured with reties_per_server(estimated)=%s, t_wait_per_server=%s, '
                    't_wait_no_tasks=%s, servers=%s', cls._pid, cls.reties_per_server, cls.t_wait_per_server,
                    cls.t_wait_no_tasks, cls.servers)

    def __init__(self):
        # we could use a metaclass or some trick on __new__ for enforcing the use of get_instance()
        if not self._thread_local.can_init:
            raise AssertionError('You must use get_instance() to get an instance')

        assert len(self.servers) > 0, 'Fatal Error: No servers have been configured'
        self._pid = os.getpid() if not self._pid else self._pid
        self._conn = None
        self._parsed_redis = parse_redis_conn(self.servers[self._active_index])

    @classmethod
    def get_instance(cls):
        if cls._thread_local.instance is None:
            cls._thread_local.can_init = True
            cls._thread_local.instance = cls()
            cls._thread_local.can_init = False
        return cls._thread_local.instance

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        if exc_type is not None:
            if issubclass(exc_type, redis.ConnectionError):
                self.mark(self.STATUS_ERROR)
                logger.error('Pid=%s, thread_id=%s ==> Lost connection to redis server: %s. Waiting %s seconds. '
                             'Exception detail follows:\n%s', self._pid, current_thread().name,
                             self.get_active_server(), self.get_wait_time(),
                             ''.join(format_exception(exc_type, exc_val, exc_tb)))
                self.wait_on_error()
                return self.__CONS_PROPAGATE_EXCEPTION

            if issubclass(exc_type, self.IdleLoopException):
                self.mark(self.STATUS_IDLE)
                logger.info("IdleLoop: %s... pid=%s, count: %s", exc_val, self._pid, self.get_message_count())
                return self.__CONS_SUPPRESS_EXCEPTION

        self.mark(self.STATUS_OK)
        return self.__CONS_PROPAGATE_EXCEPTION

    @staticmethod
    def calculate_wait_time_per_server(reties_per_server, t_wait0):
        return t_wait0 * (2**(reties_per_server + 1) - 1)

    @staticmethod
    def calculate_reties_per_server(wait_time_per_server, t_wait0):
        return int(round(math.log(wait_time_per_server * 1.0 / t_wait0 + 1, 2) - 1))

    def get_active_server(self):
        if self.should_switch_server(): #or self.should_reconnect_master():
            self.switch_active_server()  # (force_master=self.should_reconnect_master())
        return self.servers[self._active_index]

    def get_parsed_redis(self):
        return self._parsed_redis

    def should_switch_server(self):
        cur_time = time.time()
        return (self.is_previous_error() and cur_time - self._last_success_tstamp > self.t_wait_per_server) or \
               (self.is_previous_idle() and cur_time - self._last_message_tstamp > self.t_wait_no_tasks)

    def is_previous_ok(self):
        return self._retries_count == -1

    def is_previous_error(self):
        return self._retries_count > -1

    def is_previous_idle(self):
        return self._idle_count > -1

    def switch_active_server(self, force_master=False):
        self._active_index = (0 if force_master or self._active_index >= len(self.servers) - 1 else
                              self._active_index + 1)
        self._parsed_redis = parse_redis_conn(self.servers[self._active_index])

        self.mark(self.STATUS_OK)  # mark a fresh status OK for the new server

        logger.warn('Pid=%s, thread_id=%s ==> Redis Active server switched to %s, force_master=%s', self._pid,
                    current_thread().name, self.servers[self._active_index], force_master)

    def get_wait_time(self):
        return min(self.t_wait0 * (2 ** self._retries_count) if self._retries_count >= 0 and not
                   self.should_switch_server() else 0, self._max_wait_step)

    def get_message_count(self):
        return self.message_count

    def wait_on_error(self):
        time.sleep(self.get_wait_time())

    def mark(self, status):

        if status == self.STATUS_ERROR:
            self._retries_count += 1
            self._last_failure_tstamp = time.time()
            self._idle_count = -1
            self._conn = None  # force the recreation of the thread local connection in any case

        elif status == self.STATUS_IDLE:
            self._idle_count += 1
            self._retries_count = -1  # and idle loop is still a success, so clear previous errors
            self._last_success_tstamp = time.time()

        elif status == self.STATUS_OK:
            self._retries_count = -1
            self._idle_count = -1
            self._last_success_tstamp = time.time()
            self.message_count += 1
            self._last_message_tstamp = time.time()
        else:
            raise Exception('Non valid status: {}'.format(status))

    def get_healthy_conn(self):
        return self.get_conn()

    def get_conn(self):
        if self._conn is not None and not self.should_switch_server():
            return self._conn
        else:
            self._conn = None
            active_server = self.get_active_server()
            c = parse_redis_conn(active_server)
            logger.warn('Pid=%s, thread_id=%s ==> Opening new redis connection to host=%s, port=%s, db=%s', self._pid,
                        current_thread().name, c.hostname, c.port, c.virtual_host)
            self._conn = redis.StrictRedis(host=c.hostname, port=c.port, db=c.virtual_host)
            return self._conn



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import requests
from notification import BaseNotification
import logging

logger = logging.getLogger(__name__)


class HubotException(Exception):

    pass


class Hubot(BaseNotification):

    @classmethod
    def notify(cls, alert, queue, hubot_url, message=None, repeat=0):
        message = cls._get_subject(alert, custom_message=message)

        if '?' in hubot_url:
            raise ValueError

        post_params = {
            'event': queue,
            'data': message,
        }

        alert_id = alert.get('alert_def', {}).get('id', 0)

        try:
            r = requests.post(hubot_url, data=post_params)
            r.raise_for_status()
            logger.info('Notification sent: request to %s --> status: %s, response headers: %s, response body: %s',
                        hubot_url, r.status_code, r.headers, r.text)
        except Exception:
            logger.exception('Failed to send notification for alert %s with id %s to: %s', alert['name'], alert['id'],
                             hubot_url)
        finally:
            return repeat


if __name__ == '__main__':

    fake_alert = {
        'is_alert': True,
        'alert_def': {'name': 'Test'},
        'entity': {'id': 'hostxy'},
        'captures': {},
    }

    Hubot.notify(fake_alert, queue='syslog.info', hubot_url='http://z-hyp18.zalando:8081/publish')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Zalando-specific function to query the Exception Monitor
"""

from collections import Iterable

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial
from zmon_worker_monitor import plugin_manager


class ExceptionsFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(ExceptionsFactory, self).__init__()
        # fields from dependencies: plugin depends 1 other plugin
        self.http_factory = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        entity = factory_ctx['entity']
        project = entity['name'] if entity['type'] == 'project' else None

        # load plugins dependencies and store them locally for efficiency
        if not self.http_factory:
            self.http_factory = plugin_manager.get_plugin_obj_by_name('http', 'Function')

        return propartial(ExceptionsWrapper,
                          http_wrapper=self.http_factory.create(factory_ctx),
                          host=factory_ctx['host'],
                          instance=factory_ctx['instance'],
                          project=project)


class ExceptionsWrapper(object):

    def __init__(self, http_wrapper, host=None, instance=None, project=None):
        self.__http = http_wrapper
        self.url = 'https://exceptions.example.com/'
        self.host = host
        self.instance = instance
        self.project = project

    def __request(self, path, **params):
        return self.__http(path, base_url=self.url, params=params).json()

    def count(
        self,
        host=None,
        instance=None,
        project=None,
        source_class=None,
        method_name=None,
        exception_class=None,
        time_from='-24h',
        time_to=None,
        level='ERROR',
        q=None,
    ):

        return self.__request(
            'count',
            host=maybe_comma_join(host or self.host),
            instance=maybe_comma_join(instance or self.instance),
            project=maybe_comma_join(project or self.project),
            source_class=maybe_comma_join(source_class),
            method_name=maybe_comma_join(method_name),
            exception_class=maybe_comma_join(exception_class),
            time_from=time_from,
            time_to=time_to,
            level=maybe_comma_join(level),
            q=q,
        )['count']


def maybe_comma_join(s):
    """
    If `s` is iterable (but not a string), returns a comma-separated Unicode string of the elements of `s`.
    Otherwise, returns `s`

    >>> maybe_comma_join(['a', 'b', 'c'])
    u'a,b,c'

    >>> maybe_comma_join([1, 2, 3])
    u'1,2,3'

    >>> maybe_comma_join([u'\u03B1', u'\u03B2', u'\u03B3'])
    u'\u03b1,\u03b2,\u03b3'

    >>> maybe_comma_join([])
    ''

    >>> maybe_comma_join('abc')
    'abc'

    >>> maybe_comma_join(u'\u03B1\u03B2\u03B3')
    u'\u03b1\u03b2\u03b3'

    >>> maybe_comma_join('')
    ''

    >>> maybe_comma_join(123)
    123
    """

    if isinstance(s, Iterable) and not isinstance(s, basestring):
        return ','.join(unicode(e) for e in s)
    else:
        return s





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
from datetime import datetime, timedelta
import base64
import json
import logging
import time
import threading
from rpc_client import get_rpc_client
from contextlib import contextmanager
import settings
import eventloghttp
import snappy

import plugin_manager
from redis_context_manager import RedisConnHandler
from tasks import load_config_from_file, configure_tasks
from tasks import check_and_notify, trial_run, cleanup

logger = logging.getLogger(__name__)


TASK_POP_TIMEOUT = 5


__config = None


def get_config():
    global __config
    if __config is None:
        __config = settings.get_external_config() or load_config_from_file()
    return __config


def flow_simple_queue_processor(queue='', **execution_context):
    '''
    Simple logic to connect to a redis queue, listen to messages, decode them and execute the tasks

    :param queue: (str) queue to connect to
    :param execution_context: (dict) other kwargs that may have been passed when worker was spawn
    :return:

    Some info to understand celery messages:

    1. An example of a celery message as first received (base64-encoded body shortened):
    ('zmon:queue:default',
    '{
        "body": "eyJleHBpcm...t9fQ==", "headers": {}, "content-type": "application/json",
        "properties": {
            "body_encoding": "base64",
            "correlation_id": "check-277-de_zalando:access-control-kit-1409826332.92",
            "reply_to": "abc5c87f-74eb-3570-a1cf-e426eaf91ca7",
            "delivery_info": {
                "priority": 0,
                "routing_key": "default",
                "exchange": "zmon"
            },
            "delivery_mode": 2,
            "delivery_tag": "94288433-cb4e-4d33-be29-c63e2bbce39a"
        },
        "content-encoding": "utf-8"}'
    )

    2. An example of the message['body'] after being base64-decoded (args list shortened):
    {
        u'utc': True,
        u'chord': None,
        u'args': [{u'check_id': 277, u'interval': 60, u'entity': {u'instance_type': u'zomcat', ...}, u'condition': u'>100', ...}],
        u'retries': 0,
        u'expires': u'2014-09-04T10:27:32.919152+00:00',
        u'task': u'check_and_notify',
        u'callbacks': None,
        u'errbacks': None,
        u'timelimit': [90, 60],
        u'taskset': None,
        u'kwargs': {},
        u'eta': None,
        u'id': u'check-277-de_zalando:access-control-kit-1409826332.92'
    }

    '''

    known_tasks = {'check_and_notify': check_and_notify, 'trial_run': trial_run, 'cleanup': cleanup}

    #get configuration and configure tasks
    config = get_config()
    configure_tasks(config)

    logger.info('Connecting simple_queue_consumer to queue=%s, execution_context=%s', queue, execution_context)

    RedisConnHandler.configure(**dict(config))

    eventloghttp.set_target_host(config.get('eventlog.host','localhost'), config.get('eventlog.port', 8081))
    eventloghttp.enable_http(config.get('eventlog.http', True))

    reactor = FlowControlReactor.get_instance()

    conn_handler = RedisConnHandler.get_instance()

    # try cleanup captures queue in aws context
    r = conn_handler.get_healthy_conn()
    r.delete('zmon:captures2graphite')

    expired_count = 0
    count = 0

    while True:
        try:

            with conn_handler as ch:

                r_conn = ch.get_healthy_conn()

                encoded_task = r_conn.blpop(queue, TASK_POP_TIMEOUT)

                if encoded_task is None:
                    raise ch.IdleLoopException('No task received')

                queue, msg = encoded_task

                if not msg[:1] == '{':
                    msg = snappy.decompress(msg)

                msg_obj = json.loads(msg)

                msg_body = None

                body_encoding = msg_obj.get("properties", {}).get("body_encoding")
                if body_encoding == "nested":
                    msg_body = msg_obj["body"]
                elif body_encoding == "base64":
                    msg_body = json.loads(base64.b64decode(msg_obj['body']))
                elif body_encoding == "snappy":
                    msg_body = json.loads(snappy.decompress(base64.b64decode(msg_obj['body'])))

                taskname = msg_body['task']
                func_args = msg_body['args']
                func_kwargs = msg_body['kwargs']
                timelimit = msg_body.get('timelimit')  # [90, 60]
                t_hard, t_soft = timelimit

                # we pass task metadata as a kwargs right now, later will be put in the function context by our decorator
                task_context = {
                    'queue': queue,
                    'taskname': taskname,

                    'delivery_info': msg_obj.get('properties', {}).get('delivery_info', {}),
                    'task_properties': {
                        'task': taskname,
                        'id': msg_body.get('id', ''),
                        'expires': msg_body.get('expires'),  # '2014-09-04T10:27:32.919152+00:00'
                        'timelimit': timelimit,              # [90, 60]
                        'utc': msg_body.get('utc', True)
                    },
                }

                # discard tasks that are expired if expire metadata comes with the message
                cur_time = datetime.utcnow() if task_context['task_properties']['utc'] else datetime.now()
                expire_time = datetime.strptime(msg_body.get('expires').replace("Z", "").rsplit('+', 1)[0], '%Y-%m-%dT%H:%M:%S.%f') \
                    if msg_body.get('expires') else cur_time + timedelta(seconds=10)

                check_id = (msg_body['args'][0].get('check_id', 'xx') if len(msg_body['args']) > 0 and isinstance(msg_body['args'][0], dict) else 'XX')
                logger.debug('task loop analyzing time: check_id=%s, cur_time: %s , expire_time: %s, msg_body["expires"]=%s', check_id, cur_time, expire_time, msg_body.get('expires'))

                if cur_time < expire_time:
                    with reactor.enter_task_context(taskname, t_hard, t_soft):
                        known_tasks[taskname](*func_args, task_context=task_context, **func_kwargs)
                else:
                    logger.warn('Discarding task due to time expiration. cur_time: %s , expire_time: %s, msg_body["expires"]=%s  ----  msg_body=%s', cur_time, expire_time, msg_body.get('expires'), msg_body)
                    expired_count += 1
                    if expired_count % 500 == 0:
                        logger.warning("expired tasks count: %s", expired_count)
                count += 1

        except Exception:
            logger.exception('Exception in redis loop. Details: ')
            time.sleep(5) # avoid heavy log spam here
            # some exit condition on failure: maybe when number of consecutive failures > n ?

    # TODO: Clean redis connection... very important!!!!
    # disconnect_all()


def flow_forked_child(queue='', **kwargs):
    """
    Implement forking a work horse process per message as seen in python_rq
    """
    #TODO: implement
    pass


def flow_forked_childs(queue='', **kwargs):
    """
    Implement forking several work horses process per message?
    """
    #TODO: implement
    pass


def flow_multiprocessing_pool(queue='', **kwargs):
    """
    Implement spawning a pool of workers of multiprocessing process
    """
    #TODO: implement
    pass


class FlowControlReactor(object):
    """
    Implements a singleton object with a permanently running action loop, that can communicate with the
    parent process (ProcessController) to request certain actions or submit information about the health
    of this worker.
    Only implemented capability till now is a "Hard Kill" functionality that kicks in when a task is
    taking too long to complete. We use a context manager to signal when we enter or leave this mode of operations.
    Future capabilities may include periodical reports to the parent process about number of processed tasks,
    mean time spent by the N slowest running tasks. Also a soft kill feature.
    """

    _initialized = False
    _can_init = False
    _instance = None
    t_wait = 0.2

    def __init__(self):
        #self.task_agg_info = {}  # we could aggregate some info about how tasks are running in this worker
        assert not self._initialized and self._can_init, 'Call get_instance() to instantiate'
        self._initialized = True
        self._pid = os.getpid()
        self._rpc_client = get_rpc_client('http://{}:{}{}'.format(settings.RPC_SERVER_CONF['HOST'],
                                                                   settings.RPC_SERVER_CONF['PORT'],
                                                                   settings.RPC_SERVER_CONF['RPC_PATH']))
        self._current_task_by_thread = {}  # {thread_id: (taskname, t_hard, t_soft, tstart)}
        self.action_on = False
        self._thread = threading.Thread(target=self.action_loop)
        self._thread.daemon = True

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._can_init = True
            cls._instance = cls()
        return cls._instance

    @contextmanager
    def enter_task_context(self, taskname, t_hard, t_soft):
        self.task_received(taskname, t_hard, t_soft)
        try:
            yield self
        except Exception as e:
            self.task_ended(exc=e)
            raise
        else:
            self.task_ended()

    def action_loop(self):

        while self.action_on:
            try:
                # hard kill logic
                for th_name, (taskname, t_hard, t_soft, ts) in self._current_task_by_thread.copy().items():
                    if time.time() > ts + t_hard:
                        logger.warn('Hard Kill request started for worker pid=%s, task: %s, t_hard=%d', self._pid,
                                    taskname, t_hard)
                        self._rpc_client.mark_for_termination(self._pid)  # rpc call to parent asking for a kill
                        self._current_task_by_thread.pop(th_name, {})
            except Exception:
                logger.exception('Scary Error in FlowControlReactor.action_loop(): ')

            time.sleep(self.t_wait)

    def start(self):
        self.action_on = True
        self._thread.start()

    def stop(self):
        self.action_on = False

    def task_received(self, taskname, t_hard, t_soft):
        # this sets a timer for this task, there is only one task per thread, and right now only main thread produce
        self._current_task_by_thread[threading.currentThread().getName()] = (taskname, t_hard, t_soft, time.time())

    def task_ended(self, exc=None):
        # delete the task from the list
        self._current_task_by_thread.pop(threading.currentThread().getName(), {})


def start_worker_for_queue(flow='simple_queue_processor', queue='zmon:queue:default', **execution_context):
    """
    Starting execution point to the workflows
    """

    known_flows = {'simple_queue_processor': flow_simple_queue_processor}

    if flow not in known_flows:
        logger.exception("Bad role: %s" % flow)
        sys.exit(1)

    logger.info("Starting worker with pid=%s, flow type: %s, queue: %s, execution_context: %s", os.getpid(), flow,
                queue, execution_context)

    # init the plugin manager
    plugin_manager.init_plugin_manager()

    # load external plugins (should be run only once)
    plugin_manager.collect_plugins(global_config=get_config(), load_builtins=True, load_env=True)

    # start Flow Reactor here
    FlowControlReactor.get_instance().start()

    exit_code = 0
    try:

        known_flows[flow](queue=queue, **execution_context)

    except (KeyboardInterrupt, SystemExit):
        logger.warning("Caught user signal to stop consumer: finishing!")
    except Exception:
        logger.exception("Exception in start_worker(). Details: ")
        exit_code = 2
    finally:
        FlowControlReactor.get_instance().stop()
        sys.exit(exit_code)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import cherrypy

import argparse
import settings
import logging

if __name__ == '__main__':
    import logging.config
    logging.config.dictConfig(settings.RPC_SERVER_CONF['LOGGING'])

logger = logging.getLogger(__name__)

# env vars get droped via zompy startup
os.environ["ORACLE_HOME"] = "/opt/oracle/instantclient_12_1/"
os.environ["LD_LIBRARY_PATH"] = os.environ.get("LD_LIBRARY_PATH", '') + ":/opt/oracle/instantclient_12_1/"

import rpc_server

DEFAULT_NUM_PROC = 16


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config-file", help="path to config file")
    parser.add_argument('--no-rpc', action='store_true', help='Do not start XML-RPC server')
    return parser.parse_args(args)


def main(args=None):
    # add src dir to sys.path
    # src_dir = os.path.abspath(os.path.dirname(__file__))
    # if src_dir not in sys.path:
    #     sys.path.append(src_dir)

    args = parse_args(args)

    main_proc = rpc_server.MainProcess()

    # load cherrypy configuration
    if args.config_file and os.path.exists(args.config_file):
        cherrypy.config.update(args.config_file)
    elif os.path.exists('/app/web.conf'):
        cherrypy.config.update('/app/web.conf')
    else:
        cherrypy.config.update('web.conf')

    for key in cherrypy.config.keys():
        env_key = key.upper().replace('.', '_')
        if env_key in os.environ:
            cherrypy.config[key] = os.environ[env_key]

    # save cherrypy config in owr settings module
    settings.set_workers_log_level(cherrypy.config.get('loglevel', 'INFO'))
    settings.set_external_config(cherrypy.config)
    settings.set_rpc_server_port('2{}'.format('3500'))

    # start the process controller
    main_proc.start_proc_control()

    # start some processes per queue according to the config
    queues = cherrypy.config['zmon.queues']['local']
    for qn in queues.split(','):
        queue, N = (qn.rsplit('/', 1) + [DEFAULT_NUM_PROC])[:2]
        main_proc.proc_control.spawn_many(int(N), kwargs={"queue": queue, "flow": "simple_queue_processor"})

    if not args.no_rpc:
        main_proc.start_rpc_server()

    return main_proc


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Server module for exposing an rpc interface for clients to remotely control a local ProcessManager
"""

import inspect
import json
import xmlrpclib
from SimpleXMLRPCServer import SimpleXMLRPCServer
from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler

import logging
logger = logging.getLogger(__name__)


class RpcProxy(object):
    """
    This is a base class to subclass in order to expose an instance object through remote RPC
    It serves as a container for some idiosyncrasies of Python XML_RPC
    like the private methods: _listMethods , _methodHelp and _dispatch, which purpose isn't obvious at firsts.
    Here we try to take advantage of these idiosyncrasies.
    """

    exposed_obj_class = object    # override with the class of the object to expose
    valid_methods = []            # override with the list of methods you want to call from RPC

    def __init__(self, exposed_obj):
        assert type(exposed_obj) is self.exposed_obj_class, "Error in RpcProxy: exposed_obj is not declared class"
        self.exposed_obj = exposed_obj

    def _listMethods(self):
        # this method must be present for system.listMethods to work
        return self.valid_methods

    def _methodHelp(self, method):
        # Override this method for system.methodHelp to work
        if method == 'example_method':
            return "example_method(2,3) => 5"
        else:
            # By convention, return empty string if no help is available
            return ""

    def get_exposed_obj(self):
        #Never add this method to valid_methods
        return self.exposed_obj

    def on_exit(self):
        #Override this to provide a logic to be executed when server is finishing
        pass

    def signal_termination(self, terminate):
        self._signal_terminate_and_exit = bool(terminate)

    def _dispatch(self, method, params):
        #This method is automatically called by Python's SimpleXMLRPCServer for every incoming rpc call
        if method in self.valid_methods:
            obj = self  if hasattr(self, method)  else  self.exposed_obj
            try:
                kw = {}
                m = getattr(obj, method)

                if len(params) and str(params[-1]).startswith('js:'):
                    # let's try to interpret the last argument as keyword args in json format
                    _kw = json.loads(str(params[-1])[len('js:'):])
                    aspec = inspect.getargspec(m)
                    if isinstance(_kw, dict) and _kw and [k in aspec.args for k in _kw]:
                        params = params[:-1]
                        kw = _kw

                return getattr(obj, method)(*params, **kw)
            except Exception:
                logger.exception("Exception encountered in rpc_server while attempting to call: %s with params: %s ",
                                 method, params)
                raise
        else:
            raise Exception('method "%s" is not supported' % method)


def get_rpc_client(endpoint):
    """
    Get an rpc client object to remote server listening at endpoint
    :param endpoint: http://host:port/rpc_path
    :return: rpc_client object
    """
    return xmlrpclib.ServerProxy(endpoint)

#TODO: move to a method in RpcProxy
def start_RPC_server(host, port, rpc_path, rpc_proxy):
    """
    Starts the RPC server and expose some methods of rpc_proxy
    :param host:
    :param port:
    :param rpc_proxy:
    :return:
    """
    # Restrict to a particular path.
    class RequestHandler(SimpleXMLRPCRequestHandler):
        #rpc_paths = ('/RPC2',)
        rpc_paths = ('/' + rpc_path.lstrip('/'), )

    # Create server
    server = SimpleXMLRPCServer((host, port), requestHandler=RequestHandler, allow_none=True)
    server.register_introspection_functions()

    server.register_instance(rpc_proxy)

    try:
        # Run the server's main loop
        server.serve_forever()
    except (KeyboardInterrupt, SystemExit):
        logger.info("Server interrupted: Exiting!")
        rpc_proxy.on_exit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import json
import numpy

from collections import Set
from decimal import Decimal

class JsonDataEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, (datetime.datetime, datetime.date, datetime.time)):
            return o.isoformat()
        elif isinstance(o, Decimal):
            return float(o)
        elif isinstance(o, Set):
            return list(o)
        elif isinstance(o, numpy.bool_):
            return bool(o)
        else:
            return super(JsonDataEncoder, self).default(o)

    def iterencode(self, o, _one_shot=False):
        for chunk in super(JsonDataEncoder, self).iterencode(o, _one_shot=_one_shot):
            yield {'NaN': 'null', 'Infinity': '"Infinity"', '-Infinity': '"-Infinity"'}.get(chunk, chunk)


if __name__ == '__main__':

    the_encoder = JsonDataEncoder()

    print repr(the_encoder.encode(numpy.nan))
    print repr(the_encoder.encode(numpy.Infinity))
    print repr(the_encoder.encode(-numpy.Infinity))
    print repr(the_encoder.encode([numpy.nan, numpy.Infinity, -numpy.Infinity]))
    print repr(the_encoder.encode({'NaN': numpy.nan, 'Infinity': numpy.Infinity, '-Infinity': -numpy.Infinity}))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from zmon_worker_monitor.zmon_worker.errors import JmxQueryError

import json
import requests
import time

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


class JmxFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(JmxFactory, self).__init__()
        self._jmxquery_host = None
        self._jmxquery_port = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self._jmxquery_host = conf['jmxquery.host']
        self._jmxquery_port = conf['jmxquery.port']

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(JmxWrapper,
                          jmxqueryhost=self._jmxquery_host,
                          jmxqueryport=self._jmxquery_port,
                          host=factory_ctx['host'],
                          port=factory_ctx['jmx_port'])


class JmxWrapper(object):

    def __init__(self, jmxqueryhost, jmxqueryport, host, port, timeout=5):
        #jmxquery running where?
        self.jmxquery_host = jmxqueryhost
        self.jmxquery_port = jmxqueryport

        self.host = host
        self.port = port

        self.timeout = timeout
        self._queries = []

    @staticmethod
    def _transform_results(data):
        '''Transform JSON returned from JMX Query to a reasonable dict

        >>> JmxWrapper._transform_results({'results':[{'beanName':'mybean','attributes':{'HeapMemoryUsage':1}}]})
        {'HeapMemoryUsage': 1}
        >>> JmxWrapper._transform_results({'results':[{'beanName':'a','attributes':{'x':1}}, {'beanName': 'b', 'attributes': {'y': 2}}]})
        {'a': {'x': 1}, 'b': {'y': 2}}
        >>> JmxWrapper._transform_results({'results':[{'beanName':'a','attributes':{'x':{'compositeType': {}, 'contents': {'y':7}}}}]})
        {'x': {'y': 7}}
        '''

        results = data['results']
        d = {}
        for result in results:
            attr = result['attributes']

            for key, val in attr.items():
                if 'password' in key.lower():
                    attr[key] = ''
                    continue

                if isinstance(val, dict) and 'compositeType' in val and 'contents' in val:
                    # special unpacking of JMX CompositeType objects (e.g. "HeapMemoryUsage")
                    # we do not want all the CompositeType meta information => just return the actual values
                    attr[key] = val['contents']
            d[result['beanName']] = attr
        if len(d) == 1:
            # strip the top-level "bean name" keys
            return d.values()[0]
        else:
            return d

    def query(self, bean, *attributes):
        self._queries.append((bean, attributes))
        return self

    def _jmxquery_queries(self):
        for bean, attributes in self._queries:
            query = bean
            if attributes:
                query += '@' + ','.join(attributes)
            yield query

    def results(self):
        if not self._queries:
            raise ValueError('No query to execute')

        try:
            r = requests.get('http://{}:{}'.format(self.jmxquery_host, self.jmxquery_port), params={'host': self.host, 'port': self.port,
                             'query': self._jmxquery_queries()}, timeout=2)

            if r.status_code == 500:
                raise Exception('-do-one-try-')
        except:
            time.sleep(1)
            r = requests.get('http://{}:{}'.format(self.jmxquery_host, self.jmxquery_port), params={'host': self.host, 'port': self.port,
                                                                                                    'query': self._jmxquery_queries()}, timeout=2)

        output = r.text

        try:
            data = json.loads(output)
        except:
            raise JmxQueryError(output)


        return self._transform_results(data)


if __name__ == '__main__':
    # example call:
    # JAVA_HOME=/opt/jdk1.7.0_21/ python jmx.py restsn03 49600 jmxremote.password java.lang:type=Memory HeapMemoryUsage
    import sys
    jmx = JmxWrapper(*sys.argv[1:4])
    print jmx.query(*sys.argv[4:]).results()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import boto3
import collections
import datetime
import fnmatch
import json
import sys
import logging
import requests
import sys

logging.getLogger('botocore').setLevel(logging.WARN)

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial

class CloudwatchWrapperFactory(IFunctionFactoryPlugin):
    def __init__(self):
        super(CloudwatchWrapperFactory, self).__init__()

    def configure(self, conf):
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(CloudwatchWrapper, region=factory_ctx.get('entity').get('region', None))

def get_region():
    r = requests.get('http://169.254.169.254/latest/dynamic/instance-identity/document', timeout=3)
    return r.json()['region']

def matches(dimensions, filters):
    for key, pattern in filters.items():
        if not fnmatch.fnmatch(''.join(dimensions.get(key, '')), pattern):
            return False
    return True

class CloudwatchWrapper(object):

    def __init__(self, region=None):
        if not region:
            region = get_region()
        self.client = boto3.client('cloudwatch', region_name=region)

    def query(self, dimensions, metric_name, statistics='Sum', namespace=None, unit=None, period=60, minutes=5):
        filter_dimension_keys = set()
        filter_dimension_pattern = {}
        for key, val in list(dimensions.items()):
            if val == 'NOT_SET':
                filter_dimension_keys.add(key)
                del dimensions[key]
            if val and '*' in val:
                filter_dimension_pattern[key] = val
                del dimensions[key]
        dimension_kvpairs = [{'Name': k, 'Value': v} for k, v in dimensions.items()]
        args = {'Dimensions': dimension_kvpairs, 'MetricName': metric_name}
        if namespace:
            args['Namespace'] = namespace
        metrics = self.client.list_metrics(**args)
        metrics = metrics['Metrics']
        end = datetime.datetime.utcnow()
        start = end - datetime.timedelta(minutes=minutes)
        data = collections.defaultdict(int)
        data['dimensions'] = collections.defaultdict(int)
        for metric in metrics:
            metric_dimensions = {d['Name']: d['Value'] for d in metric['Dimensions']}
            if set(metric_dimensions.keys()) & filter_dimension_keys:
                continue
            if filter_dimension_pattern and not matches(metric_dimensions, filter_dimension_pattern):
                continue
            response = self.client.get_metric_statistics(Namespace=metric['Namespace'], MetricName=metric['MetricName'], Dimensions=metric['Dimensions'],
                                                         StartTime=start, EndTime=end, Period=period, Statistics=[statistics])
            data_points = response['Datapoints']
            if data_points:
                for [dim_name, dim_val] in metric_dimensions.items():
                    if not dim_name in data['dimensions']:
                        data['dimensions'][dim_name] = collections.defaultdict(int)
                    data['dimensions'][dim_name][dim_val] += data_points[-1][statistics]
                data[metric['MetricName']] += data_points[-1][statistics]
        return data


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    cloudwatch = CloudwatchWrapper(sys.argv[1])
    print "ELB result (eu-west-1 only):"
    elb_data = cloudwatch.query({'AvailabilityZone': 'NOT_SET', 'LoadBalancerName': 'pierone-*'}, 'Latency', 'Average')
    print(json.dumps(elb_data))
    print "Billing result (us-east-1 only):"
    billing_data = cloudwatch.query({'Currency': 'USD'}, 'EstimatedCharges', 'Maximum', 'AWS/Billing', None, 3600, 60*4)
    print(json.dumps(billing_data))



# -*- coding: utf-8 -*-
"""
Execution script
"""

import sys
import os
import json
import settings
import logging


def _set_logging(log_conf):
    #substituting {pid} in log handlers
    for handler_name, handler_dict in log_conf['handlers'].items():
        if '{pid}' in handler_dict.get('filename', ''):
            log_conf['handlers'][handler_name]['filename'] = log_conf['handlers'][handler_name]['filename']\
            .format(pid=os.getpid())

    import logging.config
    logging.config.dictConfig(log_conf)



def start_worker(**kwargs):
    """
    A simple wrapper for workflow.start_worker(role) , needed to solve the logger import problem with multiprocessing
    :param role: one of the constants workflow.ROLE_...
    :return:
    """
    _set_logging(settings.LOGGING)

    import workflow

    workflow.start_worker_for_queue(**kwargs)


if __name__ == '__main__':

    if len(sys.argv) != 2:
        print >> sys.stderr, "\nWrong command line parameters:\n  " \
                             "usage: {0} <kwargs_json>".format(sys.argv[0])
        sys.exit(1)

    kwargs = json.loads(sys.argv[1])

    print "unmanaged worker started with kwargs: {}".format(kwargs)

    start_worker(**kwargs)
